"""
Module containing the FooWrapper1 class
"""
import time

import subprocess

from wopmars.framework.database.tables.ToolWrapper import ToolWrapper


class FooWrapper9(ToolWrapper):
    """
    This class has been done for example/testing purpose.
    Modifications may lead to failure in tests.
    """
    __mapper_args__ = {'polymorphic_identity': "FooWrapper9"}

    def specify_input_file(self):
        return ["input1"]

    def specify_input_table(self):
        return ["FooBase2"]

    def specify_output_file(self):
        return ["output1"]

    def run(self):
        print(self.__class__.__name__ + " is running...")
        p = subprocess.Popen(["touch", self.output_file("output1")])
        p.wait()
        time.sleep(1)



import os
import unittest
from unittest import TestCase

from sqlalchemy.orm.exc import MultipleResultsFound

from test.resource.model.FooBase import FooBase
from wopmars.framework.database.SQLManager import SQLManager
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.PathFinder import PathFinder


class TestWopMarsSession(TestCase):
    def setUp(self):
        OptionManager.initial_test_setup()
        self.s_root_path = PathFinder.get_module_path()
        SQLManager.instance().create_all()
        self.__session = SQLManager.instance().get_session()

    def tearDown(self):
        self.__session.rollback()
        SQLManager.instance().get_session().close()
        SQLManager.instance().drop_all()
        PathFinder.dir_content_remove(os.path.join(self.s_root_path, "test/output"))
        OptionManager._drop()
        SQLManager._drop()

    def test_commit_query_add(self):
        for i in range(10):
            f = FooBase(name="testSession " + str(i))
            self.__session.add(f)
        self.__session.commit()
        self.assertEqual(len(self.__session.query(FooBase).all()), 10)

    def test_rollback(self):
        for i in range(10):
            f = FooBase(name="testSession " + str(i))
            self.__session.add(f)
        self.__session.rollback()
        self.assertEqual(len(self.__session.query(FooBase).all()), 0)

    def test_add_all(self):
        fs = []
        for i in range(10):
            fs.append(FooBase(name="testSession " + str(i)))
        self.__session.add_all(fs)
        self.__session.commit()
        self.assertEqual(len(self.__session.query(FooBase).all()), 10)

    def test_delete(self):
        fs = []
        for i in range(10):
            fs.append(FooBase(name="testSession " + str(i)))
        self.__session.add_all(fs)
        self.__session.commit()

        for i in range(5):
            f = self.__session.query(FooBase).filter(FooBase.name == "testSession " + str(i)).first()
            self.__session.delete(f)
        self.assertEqual(len(self.__session.query(FooBase).all()), 5)

    def test_delete_content(self):
        fs = []
        for i in range(10):
            fs.append(FooBase(name="testSession " + str(i)))
        self.__session.add_all(fs)
        self.__session.commit()

        self.__session.delete_content(FooBase)
        self.assertEqual(len(self.__session.query(FooBase).all()), 0)

    def test_something(self):
        fs = []
        for i in range(10):
            fs.append(FooBase(name="testSession " + str(i)))
        self.__session.add_all(fs)
        self.assertTrue(self.__session.something())

        self.__session.commit()
        self.assertFalse(self.__session.something())

        for i in range(10):
            self.__session.query(FooBase).filter(FooBase.name == "testSession " + str(i)).first().name = "sessionTest " + str(i)
        self.assertTrue(self.__session.something())

    def test_get_or_create(self):
        fs = []
        for i in range(10):
            fs.append(FooBase(name="testSession " + str(i)))
        self.__session.add_all(fs)
        self.__session.commit()
        self.assertEqual(self.__session.get_or_create(FooBase, name="testSession 0")[0], fs[0])
        self.__session.delete(fs[0])
        self.assertNotEqual(self.__session.get_or_create(FooBase, name="testSession 0")[0], fs[0])

    def test_query(self):
        fs = []
        for i in range(10):
            fs.append(FooBase(name="testSession " + str(i)))
        self.__session.add_all(fs)
        self.__session.commit()

        self.assertEqual(len(self.__session.query(FooBase).all()), 10)

        with self.assertRaises(MultipleResultsFound):
            self.assertEqual(self.__session.query(FooBase).one().name, "testSession 0")

        self.assertEqual(self.__session.query(FooBase).filter(FooBase.name == "testSession 0").one().name, "testSession 0")

        self.assertEqual(self.__session.query(FooBase).count(), 10)

        self.assertEqual(self.__session.query(FooBase).first().name, "testSession 0")

        self.assertIsNone(self.__session.query(FooBase).filter(FooBase.name == "existepas").one_or_none())

        self.assertIsNone(self.__session.query(FooBase).filter(FooBase.name == "existepas").scalar())
        self.assertEqual(self.__session.query(FooBase.name).filter(FooBase.name == "testSession 0").scalar(), "testSession 0")
        with self.assertRaises(MultipleResultsFound):
            self.assertEqual(self.__session.query(FooBase.name).scalar(), "testSession 0")

if __name__ == '__main__':
    unittest.main()



from unittest import TestCase

import os
import subprocess
import time
import unittest

from wopmars import OptionManager
from wopmars.framework.database.SQLManager import SQLManager
from wopmars.framework.database.tables.Execution import Execution
from wopmars.utils.PathFinder import PathFinder
from wopmars import WopMars

from wopmars.constants import home_wopmars


class TestWopMars(TestCase):


    def setUp(self):
        if not os.path.isdir(home_wopmars):
            os.makedirs(home_wopmars)
        self.s_root_path = PathFinder.get_module_path()
        os.chdir(self.s_root_path)
        if 'DB_URL' not in os.environ:
            self.__db_url = "sqlite:///" + os.path.join(self.s_root_path, "test/output/db.sqlite")
        else:
            self.__db_url = os.environ['DB_URL']
        #
        self.__example_def_file1 = os.path.join(self.s_root_path, "test/resource/wopfile/example_def_file1.yml")
        self.__example_def_file4 = os.path.join(self.s_root_path, "test/resource/wopfile/example_def_file4.yml")
        self.__example_def_file2_only_files = os.path.join(self.s_root_path, "test/resource/wopfile/example_def_file2_only_files.yml")
        self.__example_def_file5_never_ready = os.path.join(self.s_root_path, "test/resource/wopfile/example_def_file5_never_ready.yml")
        self.__example_def_file_input_not_ready = os.path.join(self.s_root_path, "test/resource/wopfile/example_def_file_input_not_ready.yml")

    def test_01run(self):
        cmd_line = ["python", "-l", "-D", self.__db_url, "-w", self.__example_def_file1, "-v", "-p", "-d", PathFinder.get_module_path()]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        self.assertTrue(os.path.exists(os.path.join(self.s_root_path, 'test/output/output_file1.txt')))
        self.assertTrue(os.path.exists(os.path.join(self.s_root_path, 'test/output/output_file2.txt')))
        self.assertTrue(os.path.exists(os.path.join(self.s_root_path, 'test/output/output_file7.txt')))
        self.assertEqual(se.exception.code, 0)

    def test_02dry_run(self):
        cmd_line = ["python", "-n", "-D", self.__db_url, "-w", self.__example_def_file1, "-vv", "-p", "-d",
                    PathFinder.get_module_path()]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        self.assertEqual(se.exception.code, 0)

    def test_03run_that_fail(self):
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file5_never_ready, "-vv", "-p", "-d",
                    PathFinder.get_module_path()]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        self.assertFalse(os.path.exists(os.path.join(self.s_root_path, 'resources/never_done.txt')))
        self.assertEqual(se.exception.code, 1)

    def test_04run_sourcerule_fail(self):
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-vv", "-p", "--sourcerule", "failure"]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        self.assertFalse(os.path.exists(os.path.join(self.s_root_path, 'test/output/output_file1.txt')))
        self.assertEqual(se.exception.code, 1)

    def test_05run_sourcerule_succeed(self):
        p=subprocess.Popen(["touch", os.path.join(self.s_root_path, "test/output/output_file1.txt")])
        p.wait()
        p=subprocess.Popen(["touch", os.path.join(self.s_root_path, "test/output/output_file2.txt")])
        p.wait()
        p=subprocess.Popen(["touch", os.path.join(self.s_root_path, "test/output/output_file3.txt")])
        p.wait()
        p=subprocess.Popen(["touch", os.path.join(self.s_root_path, "test/output/output_file4.txt")])
        p.wait()
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-d", PathFinder.get_module_path(), "-vv", "-p", "--sourcerule", "rule2"]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        self.assertEqual(se.exception.code, 0)

    def test_06run_skipping_steps_time_check(self):
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file2_only_files, "-vv", "-p"]
        start = time.time()
        with self.assertRaises(SystemExit):
           WopMars().run(cmd_line)
        end = time.time()
        runtime1 = end - start
        start = time.time()
        with self.assertRaises(SystemExit):
           WopMars().run(cmd_line)
        end = time.time()
        runtime2 = end - start
        self.assertGreater(runtime1 * 1.5, runtime2)
        PathFinder.silentremove("test/output/output_file1.txt")

    def test_07dry_drun_skipping(self):
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file2_only_files, "-vv", "-p"]
        with self.assertRaises(SystemExit):
           WopMars().run(cmd_line)

        with self.assertRaises(SystemExit):
           WopMars().run(cmd_line + ["-n"])

    def test_08dry_drun_skipping_all_but_one(self):
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-vv", "-p"]
        with self.assertRaises(SystemExit):
           WopMars().run(cmd_line)
        PathFinder.silentremove("test/output/output_file7.txt")
        with self.assertRaises(SystemExit):
           WopMars().run(cmd_line + ["-n"])

    def test_09run6(self):
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-p", "-vv"]
        start = time.time()
        with self.assertRaises(SystemExit):
           WopMars().run(cmd_line)
        end = time.time()
        runtime1 = end - start
        start = time.time()
        with self.assertRaises(SystemExit):
           WopMars().run(cmd_line)
        end = time.time()
        runtime2 = end - start
        self.assertGreater(runtime1 * 1.5, runtime2)
        # SQLManager.instance().drop_all()
        # OptionManager._drop()
        # SQLManager._drop()
        # PathFinder.silentremove("test/output/output_file1.txt")
        # start = time.time()
        # with self.assertRaises(SystemExit):
        #    WopMars().run(cmd_line)
        # end = time.time()
        # runtime2 = end - start
        # self.assertTrue(runtime1 * 0.4 <= runtime2 <= runtime1 * 1.4)

    def get_best_factor(self, full_exec_time, rule_count, maximum_ratio=1):
        average_rule_time = full_exec_time / rule_count
        return 1 + (maximum_ratio * average_rule_time / (1.5 + average_rule_time))

    def test_10run_error_in_command_line(self):
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "--dot", "/usr/"]
        with self.assertRaises(SystemExit) as SE:
            WopMars().run(cmd_line)
        self.assertEqual(SE.exception.code, 2)

        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "--not-known"]
        with self.assertRaises(SystemExit) as SE:
            WopMars().run(cmd_line)
        self.assertEqual(SE.exception.code, 2)

    def test_run_packaged_wrappers(self):
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file4, "-vv", "-p", "-d",
                    PathFinder.get_module_path()]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        self.assertEqual(se.exception.code, 0)

    def test_run_one_tool(self):
        cmd_line = ["python", "tool", "test.resource.wrapper.FooWrapper4",
                  "-i", "{'file': {'input1': 'test/resource/input_files/input_file1.txt'}}",
                  "-o", "{'file': {'output1': 'test/output/output1.txt'}}", "-D", self.__db_url, "-vv", "-p", "-d",
                  PathFinder.get_module_path()]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        self.assertEqual(se.exception.code, 0)

    def test_core(self):
        cmd_line = ["python", "tool", "test.resource.wrapper.FooWrapperCore",
                  "-o", "{'table': {'FooBase7': 'test.resource.model.FooBase7'}}",
                  "-vv", "-p", "-D", self.__db_url, "-d", PathFinder.get_module_path()]
        with self.assertRaises(SystemExit) as se:
           WopMars().run(cmd_line)
        self.assertEqual(se.exception.code, 0)

    def test_clear_history(self):
        cmd_line = ["python", "-D", self.__db_url,
                  "-w", self.__example_def_file1,
                  "-vv", "-p", "-d", PathFinder.get_module_path(), "-c", "-F"]
        with self.assertRaises(SystemExit):
               WopMars().run(cmd_line)
               WopMars().run(cmd_line)
        session = SQLManager.instance().get_session()
        self.assertEqual(session.query(Execution).count(), 1)

    # def test_pandas(self):
    #     cmd_line = ["python", "tool", "test.resource.wrapper.FooWrapperDataframe",
    #                 "-o", "{'table': {'FooBase': 'test.resource.model.FooBase'}}",
    #                 "-vv", "-p", "-D", self.__db_url, "-d", PathFinder.get_module_path()]
    #
    #     with self.assertRaises(SystemExit) as se:
    #         WopMars().run(cmd_line)
    #
    #     self.assertEqual(se.exception.code, 1)

    def test_run_target_rule(self):
        # SQLManager.instance().create_all()
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-vv", "-p", "--targetrule", "rule3"]
        with self.assertRaises(SystemExit) as se:
           WopMars().run(cmd_line)
        self.assertEqual(se.exception.code, 0)

    def test_run_input_file_not_ready(self):
        # SQLManager.instance().create_all()
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file_input_not_ready, "-vv", "-p"]
        with self.assertRaises(SystemExit) as se:
           WopMars().run(cmd_line)
        self.assertEqual(se.exception.code, 1)

    def tearDown(self):
        SQLManager.instance().get_session().close()
        SQLManager.instance().drop_all()
        PathFinder.dir_content_remove(os.path.join(self.s_root_path, "test/output"))
        OptionManager._drop()
        SQLManager._drop()

if __name__ == "__main__":
    unittest.main()





"""
Example of module documentation which can be
multiple-lined
"""
import errno
import importlib
import os

from wopmars.utils.exceptions.WopMarsException import WopMarsException

class PathFinder:
    """
    Static class for finding paths
    """
    @staticmethod
    def get_module_path():
        """
        Find the Src directory of the project

        :return: the path leading to the src file of the project
        """

        module_path = os.path.join(os.path.dirname(__file__), "../..")
        return module_path

    @staticmethod
    def check_pygraphviz(path):
        importlib.import_module("pygraphviz")

    @staticmethod
    def create_workingdir(path):
        try:
            os.makedirs(path)
            return True
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise
                return True

    @staticmethod
    def check_database_valid_url(url):
        """
        Check if the path given is correct on the system.

        If path is None, return None.

        :param path:
        :return:
        """
        db_connection = url.split("://")[0]
        if db_connection == "sqlite":
            sqlite_db_path=url.replace("sqlite:///", "")
            if not os.path.isabs(sqlite_db_path):
                sqlite_db_path = os.path.join(os.getcwd(), sqlite_db_path)
            try:
                os.makedirs(os.path.dirname(sqlite_db_path))
            except OSError as exception:
                if exception.errno != errno.EEXIST:
                    raise
            if sqlite_db_path is None or os.access(os.path.dirname(os.path.abspath(os.path.expanduser(sqlite_db_path))), os.W_OK) or sqlite_db_path[0] == "$":
                return sqlite_db_path
            else:
                raise FileNotFoundError

    @staticmethod
    def check_valid_path(path):
        """
        Check if the path given is correct on the system.

        If path is None, return None.

        :param path:
        :return:
        """
        if path is None or os.access(os.path.dirname(os.path.abspath(os.path.expanduser(path))), os.W_OK) or path[0] == "$":
            return path
        else:
            raise FileNotFoundError

    @staticmethod
    def silentremove(path):
        """
        Remove a file that may not exist.
        :param path:
        :return:
        """
        try:
            os.remove(path)
        except OSError:
            pass

    @staticmethod
    def dir_content_remove(path):
        for f in os.listdir(path):
            if not f.startswith("."):
                PathFinder.silentremove(os.path.join(path, f))

    @staticmethod
    def is_in_python_path(name):
        print("dfsdfqsfsqfsqfqsdf")
        importlib.import_module(name)



from wopmars.framework.database.tables.ToolWrapper import ToolWrapper


class AddWithOrWithoutDeleteBefore(ToolWrapper):
    __mapper_args__ = {'polymorphic_identity': "sprintFive.AddWithOrWithoutDeleteBefore"}

    def specify_input_file(self):
        return ["input1"]

    def specify_output_table(self):
        return ["FooBase", "FooBase2"]

    def run(self):
        self.session().delete_content(self.output_table("FooBase"))
        for i in range(100):
            self.session().add(self.output_table("FooBase")(name="DeleteThenAdd" + str(i)))
            self.session().add(self.output_table("FooBase2")(name="Add.py" + str(i)))

        self.session().commit()



from wopmars.framework.database.Base import Base

from wopmars.utils.exceptions.WopMarsException import WopMarsException
from sqlalchemy import Column, Integer, String, ForeignKey
from sqlalchemy.orm import relationship


class Option(Base):
    """
    The Option class handle a key and a value and is able to say if it is properly formated. It is also the model of the
    table ``wom_option`` which contains the following fields:

    - id: INTEGER - primary key - auto increment
    - name: VARCHAR(255) - the name of the referenced option
    - value: VRACHAR(255) - the value of the option
    - rule_id: INTEGER - the ID of the associated rule
    """

    # static value necessary to perform tests on Options
    static_option_castable = (
        "bool",
        "str",
        "int",
        "float"
       )
    static_option_req = "required"
    static_option_default = "optional"

    __tablename__ = "wom_option"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(255))
    value = Column(String(255))
    rule_id = Column(Integer, ForeignKey("wom_rule.id"))

    rule = relationship("ToolWrapper", back_populates="options", enable_typechecks=False)

    def correspond(self, carac):
        """
        Check if the option value correspond to the type given by the tool wrapper. Throws a WopMarsException if
        not.

        :raise: WopMarsException
        :param carac: String containing the carac of the option in the format: "carac1|carac2|carc3"
        :return:
        """
        # get a list of caracteristics
        list_splitted_carac = carac.split("|")
        for s_type in list_splitted_carac:
            s_formated_type = s_type.strip().lower()
            # check if the carac is a autorized castable type
            if s_formated_type in Option.static_option_castable:
                try:
                    # try the cast
                    eval(s_formated_type)(self.value)
                except ValueError:
                    # if it fails, raise an exception: the type has not been respected
                    raise WopMarsException("The content of the definition file is not valid.",
                                           "The given option value of " + str(self.name) +
                                           " should be of type " + s_formated_type)

            # if not, it may be "default" or "required"
            else:
                # we check if an option is required and has no default value
                if s_formated_type != Option.static_option_default and s_formated_type != Option.static_option_req:
                    raise WopMarsException("Malformed toolwrapper class.",
                                           "The toolwrapper " + str(self.rule.toolwrapper) + " of the rule " +
                                           str(self.rule.name) + " has an incorrect \"specify_params\" method wich is " +
                                           "associating the " + self.name + " option with an unknown type. " +
                                           "Found: " + s_type + " - Allowed: " +
                                           str(",".join(Option.static_option_castable)))

    def __eq__(self, other):
        return self.value == other.value and self.name == other.name

    def __hash__(self):
        return id(self)

    def __str__(self):
        return str(self.name) + ": " + str(self.value)



from wopmars.framework.database.tables.ToolWrapper import ToolWrapper


class AddH(ToolWrapper):
    __mapper_args__ = {'polymorphic_identity': "sprintFive.AddH"}

    def specify_output_table(self):
        return ["FooBaseH"]

    def specify_params(self):
        return {"rows": "required|int", "del": "bool"}

    def run(self):
        if self.option("del"):
            self.session().delete_content(self.output_table("FooBaseH"))
        for i in range(self.option("rows")):
            self.session().add(self.output_table("FooBaseH")(name="Add" + str(i)))

        self.session().commit()



from wopmars.framework.database.Base import Base

from sqlalchemy import Column, String, Integer
from sqlalchemy import UniqueConstraint

class Phenotype(Base):
    __tablename__ = 'Phenotype'
    __table_args__ = (
            )

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(255), nullable=False, unique=True)




"""
Example of module documentation which can be
multiple-lined
"""
from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import ForeignKey

from wopmars.framework.database.Base import Base


class FooBase3(Base):
    """
    Documentation for the class
    """
    __tablename__ = "FooBase3"

    id = Column(Integer, primary_key=True)
    id_foobase6 = Column(Integer, ForeignKey("FooBase6.id"))
    name = Column(String(255))

    foobase4 = relationship("FooBase4", back_populates="foobase3")
    foobase6 = relationship("FooBase6", back_populates="foobase3")



import datetime
import os

import time
from sqlalchemy import Column, Integer, String, ForeignKey, DateTime, Float
from sqlalchemy.orm import relationship

from wopmars.framework.database.Base import Base
from wopmars.framework.database.SQLManager import SQLManager
from wopmars.framework.database.tables.Execution import Execution
from wopmars.framework.database.tables.Option import Option
from wopmars.utils.Logger import Logger
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.exceptions.WopMarsException import WopMarsException


class ToolWrapper(Base):
    """
    The class ToolWrapper is the superclass of the wrappers which will be designed by the wrapper developers. It is the
    SQLAlchemy Model of the table ``wom_rule`` with the following fields:

    - id: INTEGER - primary_key - auto increment - arbitrary ID
    - name: VARCHAR(255) - the name of the rule
    - toolwrapper: VARCHAR(255) - the name of the Toolwrapper
    - execution_id: INTEGER - foreign key to the table ``wom_execution`` - the associated execution
    - started_at: DATE - date at wich the toolwrapper started its execution
    - finished_at: DATE - date at wich the toolwrapper finished its execution
    - time: FLOAT - the total time toolwrapper execution
    - status: VARCHAR(255) - the final status of the Toolwrapper. it can be:

       - NOT PLANNED: the toolwrapper execution wasn't evene xpected by the user
       - ALREADY EXECUTED: the toolwrapper has been previously executed in an old workflow and doesn't need to be re-executed
       - EXECUTED: the toolwrapper has been executed
       - EXECUTION_ERROR: the toolwrapper has encountered an error during the execution
    """

    __tablename__ = "wom_rule"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(255))
    toolwrapper = Column(String(255))
    execution_id = Column(Integer, ForeignKey("wom_execution.id"))
    started_at = Column(DateTime, nullable=True)
    finished_at = Column(DateTime, nullable=True)
    time = Column(Float, nullable=True)
    status = Column(String(255), nullable=True, default="NOT_EXECUTED")

    # One rule has Many table
    tables = relationship("IODbPut", back_populates="rule")
    # One rule has Many file
    files = relationship("IOFilePut", back_populates="rule")
    # One rule has Many option
    options = relationship("Option", back_populates="rule")
    # One rule has one execution
    execution = relationship("Execution", back_populates="rules")

    # parentrules = relationship etc...
    __mapper_args__ = {
        'polymorphic_on': toolwrapper,
        'polymorphic_identity': 'ToolWrapper'
    }

    NEW = 1
    READY = 2
    NOT_READY = 3

    def __init__(self, rule_name=""):
        """
        The constructor of the toolwrapper, must not be overwritten.

        self.__state is the state given to the Toolwrapper to let the
        :class:`~.wopmars.framework.management.WorflowManager.WorkflowManager` knows if the Toolwrapper is
        able to be executed or not.

        self.__session is the session (WopMarsSession) associated with the Toolwrapper and which will be used in the run method.
        self.__state is an integer which says the actual state of the TooLWrapper: ``NEW``, ``READY``, ``NOT_READY``

        :param rule_name: the name of the rule
        :type rule_name: str
        """
        super().__init__(name=rule_name)
        self.__state = ToolWrapper.NEW
        self.__session = None

    ### PARSING METHODS

    def is_content_respected(self):
        """
        Parsing method:

        This method checks if the parameters dictionary are properly formed, according to the specifications of the
        wrapper developer.

        Call of the methods:

        - :meth:`~.wopmars.framework.database.ToolWrapper.ToolWrapper.is_options_respected`
        - :meth:`~.wopmars.framework.database.ToolWrapper.ToolWrapper.is_input_respected`
        - :meth:`~.wopmars.framework.database.ToolWrapper.ToolWrapper.is_output_respected`
        """
        # the options have to be checked first because they can alter the behavior of the is_input_respected and
        # is_output_respected methods
        self.is_options_respected()

        self.is_input_respected()
        self.is_output_respected()

    def is_input_respected(self):
        """
        Parsing method:

        Check if the input file variables names associated with the toolwrapper are ok according to the toolwrapper developer.

        It checks if the input variable names exists or not. If not, throws a WopMarsParsingException.

        This method calls the :meth:`~.wopmars.framework.database.ToolWrapper.ToolWrapper.specify_input_file` method
        which have been written by the toolwrapper developer.

        :raise WopMarsException: The input are not respected by the user.
        """
        set_input_file_names = set([f_input.name for f_input in self.files if f_input.type.name == "input"])
        # check if the input file names for the ToolWrapper are coherent with the ToolWrapper specifications
        if set_input_file_names != set(self.specify_input_file()):
            raise WopMarsException("The content of the definition file is not valid.",
                                   "The given input file variable names for " + self.__class__.__name__ +
                                   " (rule " + str(self.name) + ")" +
                                   " are not correct, they should be: " +
                                   "\n\t'{0}'".format("'\n\t'".join(self.specify_input_file())) +
                                   "\n" + "They are:" +
                                   "\n\t'{0}'".format("'\n\t'".join(set_input_file_names))
                                   )

        set_input_table = set([t_input for t_input in self.tables if t_input.type.name == "input"])
        set_input_table_names = set([t_input.tablename for t_input in set_input_table])

        # check if the input table names for the ToolWrapper are coherent with the ToolWrapper specifications
        # this condition may be a duplicate... # todo to fix?
        if set_input_table_names != set(self.specify_input_table()):
            raise WopMarsException("The content of the definition file is not valid.",
                                   "The given input table variable names for " + self.__class__.__name__ +
                                   " (rule " + str(self.name) + ")" +
                                   " are not correct, they should be: " +
                                   "\n\t'{0}'".format("'\n\t'".join(self.specify_input_table())) +
                                   "\n" + "They are:" +
                                   "\n\t'{0}'".format("'\n\t'".join(set_input_table_names))
                                   )

        for t_input in set_input_table:
            s_tablename = t_input.tablename
            if s_tablename not in self.specify_input_table():
                raise WopMarsException("The content of the definition file is not valid.",
                                       "The given input tablenames for " + 
                                       self.__class__.__name__ + " (rule " + str(self.name) + ")" +
                                       " is not correct. it should be in: " +
                                       "\n\t'{0}'".format("'\n\t'".join(self.specify_input_table())) +
                                       "\n" + "It is:" +
                                       "\n\t'" + s_tablename
                                       )
            s_tablename_of_model = t_input.get_table().__tablename__
            if s_tablename_of_model not in self.specify_input_table():
                raise WopMarsException("The content of the definition file is not valid.",
                                       "The given tablename of model for " +
                                       self.__class__.__name__ +
                                       " (rule " + str(self.name) + ")" +
                                       " is not correct. it should be in: " +
                                       "\n\t'{0}'".format("'\n\t'".join(self.specify_input_table())) +
                                       "\n" + "It is:" +
                                       "\n\t'" + s_tablename_of_model
                                       )

    def is_output_respected(self):
        """
        Parsing method:

        Check if the output dictionary given in the constructor is properly formed for the tool.

        It checks if the output variable names exists or not. If not, throws a WopMarsParsingException.

        This method calls the "specify_output_file" method which have been written by the toolwrapper developer.

        :raises WopMarsException: The output are not respected by the user.
        """
        if set([f_output.name for f_output in self.files if f_output.type.name == "output"]) != set(self.specify_output_file()):
            raise WopMarsException("The content of the definition file is not valid.",
                                   "The given output variable names for " + self.__class__.__name__ +
                                   " (rule " + str(self.name) + ")" +
                                   " are not correct, they should be: " +
                                   "\n\t'{0}'".format("'\n\t'".join(self.specify_output_file())) +
                                   "\n" + "They are:" +
                                   "\n\t'{0}'".format("'\n\t'".join([f.name for f in self.files if f.type.name == "output"]))
                                   )

        set_output_table = set([t_output for t_output in self.tables if t_output.type.name == "output"])
        set_output_table_names = set([t_input.tablename for t_input in set_output_table])
        if set_output_table_names != set(self.specify_output_table()):
            raise WopMarsException("The content of the definition file is not valid.",
                                   "The given output table variable names for " + self.__class__.__name__ +
                                   " (rule " + str(self.name) + ")" +
                                   " are not correct, they should be: " +
                                   "\n\t'{0}'".format("'\n\t'".join(self.specify_output_table())) +
                                   "\n" + "They are:" +
                                   "\n\t'{0}'".format("'\n\t'".join(set_output_table_names))
                                   )
        for t_output in set_output_table:
            s_tablename = t_output.tablename
            if s_tablename not in self.specify_output_table():
                raise WopMarsException("The content of the definition file is not valid.",
                                       "The given output tablenames for " + 
                                       self.__class__.__name__ + " (rule " + str(self.name) + ")" +
                                       " is not correct. it should be in: " +
                                       "\n\t'{0}'".format("'\n\t'".join(self.specify_output_table())) +
                                       "\n" + "It is:" +
                                       "\n\t'" + s_tablename
                                       )   

    def is_options_respected(self):
        """
        Parsing method:

        This method check if the params given in the constructor are properly formed for the tool.

        It checks if the params names given by the user exists or not, if the type correspond and if the required
        options are given. If not, throws a WopMarsParsingException.

        This method calls the "specify_params" method of the toolwrapper. This method should return a dictionnary
        associating the name of the option with a String containing the types allowed with it. A "|" is used between
        each types allowed for one option.

        Example:

        .. code-block:: python

            {
                'option1': "int",
                'option2': "required|str",
            }

        :raises WopMarsException: If the params names and types are not respected by the user.
        """
        dict_wrapper_opt_carac = self.specify_params()

        # check if the given options are authorized
        if not set([opt.name for opt in self.options]).issubset(dict_wrapper_opt_carac):
            raise WopMarsException("The content of the definition file is not valid.",
                                   "The given option variable for the rule " + str(self.name) + " -> " + self.__class__.__name__ +
                                   " are not correct, they should be in: " +
                                   "\n\t'{0}'".format("'\n\t'".join(dict_wrapper_opt_carac)) +
                                   "\n" + "They are:" +
                                   "\n\t'{0}'".format("'\n\t'".join([opt.name for opt in self.options]))
                                   )

        # check if the types correspond
        for opt in self.options:
            opt.correspond(dict_wrapper_opt_carac[opt.name])

        # check if the required options are given
        for opt in dict_wrapper_opt_carac:
            if "required" in str(dict_wrapper_opt_carac[opt]).lower() and opt not in [opt2.name for opt2 in self.options]:
                raise WopMarsException("The content of the definition file is not valid.",
                                       "The option '" + opt + "' has not been provided but it is required.")

    def follows(self, other):
        """
        Parsing method:

        Check whether the "self" follows directly "other" in the execution DAG.

        Check whether "other" has one output value in "self" possible input values.
        The output value are given from the Relationnal mapping between toolwrappers and related objects:
        * IODbPut
        * IOFilePut

        :param other: ToolWrapper that is possibly a predecessor of "self"
        :return: bool True if "self" follows "other"
        """
        for rule_f_path in [f.path for f in self.files if f.type.name == "input"]:
            for rule_f2_path in [f.path for f in other.files if f.type.name == "output"]:
                if rule_f_path == rule_f2_path:
                    return True

        for rule_t_name in [t.model for t in self.tables if t.type.name == "input"]:
            for rule_t2_name in [t.model for t in other.tables if t.type.name == "output"]:
                if rule_t_name == rule_t2_name:
                    return True

        return False

    ### Workflow Manager methods

    def get_input_files_not_ready(self):
        """
        Check if inputs are ready

        :return: bool - True if inputs are ready.
        """
        input_files_not_ready = []
        input_files = [f for f in self.files if f.type.name == "input"]
        for i in input_files:
            if not i.is_ready():
                input_files_not_ready.append(i)
        return input_files_not_ready

    def are_inputs_ready(self):
        """
        Check if inputs are ready

        :return: bool - True if inputs are ready.
        """
        input_files = [f for f in self.files if f.type.name == "input"]
        Logger.instance().debug("Inputs files of " + str(self.__class__.__name__) + ": " + str([i.name for i in input_files]))
        for i in input_files:
            if not i.is_ready():
                Logger.instance().debug("Input: " + str(i.name) + " is not ready.")
                self.__state = ToolWrapper.NOT_READY
                return False
            Logger.instance().debug("Input: " + str(i.name) + " is ready.")

        input_tables = [t for t in self.tables if t.type.name == "input"]
        Logger.instance().debug("Inputs tables of " + str(self.__class__.__name__) + ": " + str([i.tablename for i in input_tables]))
        for i in input_tables:
            if not i.is_ready():
                Logger.instance().debug("Input: " + str(i.tablename) + " is not ready.")
                self.__state = ToolWrapper.NOT_READY
                return False
            Logger.instance().debug("Input: " + str(i.tablename) + " is ready.")

        self.__state = ToolWrapper.READY
        return True

    def set_args_date_and_size(self, type, dry=False):
        """
        WorkflowManager method:

        The date and the size of the files are set according to the actual date of last modification and size of the system files

        The date of the tables are set according to the date of last modification notified in the modification_table table
        If the type of IOPut is "output" and the execution is "not dry", the date in modification_table is set to the
        current time.time() datetime.

        # todo modify it to take commits into account isntead of the status of 'output' of a table

        :param type: "input" or "output"
        :type type: str
        :param dry: Say if the execution has been simulated.
        :type dry: bool
        """
        session = SQLManager.instance().get_session()
        for f in [f for f in self.files if f.type.name == type]:
            try:
                date = datetime.datetime.fromtimestamp(os.path.getmtime(f.path))
                size = os.path.getsize(f.path)
            except FileNotFoundError as FE:
                # todo ask lionel sans ce rollback, ca bug, pourquoi? la session est vide... comme si la query etait bloquante
                if not OptionManager.instance()["--dry-run"]:
                    session.rollback()
                    raise WopMarsException("Error during the execution of the workflow",
                                           "The " + type + " file " + str(f.path) + " of rule " + str(self.name) +
                                           " doesn't exist")
                else:
                    # in dry-run mode, input/output files might not exist
                    date = None
                    size = None
            f.used_at = date
            f.size = size
            session.add(f)
            if type == "input":
                Logger.instance().debug("Input file " + str(f) + " used.")
            elif type == "output" and dry:
                Logger.instance().debug("Output file " + str(f) + " has been loaded from previous execution.")
            elif type == "output" and not dry:
                Logger.instance().debug("Output file " + str(f) + " has been created.")
        # this commit is due to a bug that i couldn't figure out: the session empty itself between the two loops...
        # this is not good at all since it may lead to inconsistence in the database
        session.commit()

        for t in [t for t in self.tables if t.type.name == type]:
            t.used_at = t.modification.date
            session.add(t)
        session.commit()

    def same_input_than(self, other):
        """
        Never used.

        Check if the other ToolWrapper have the same input than self.

        The input are say "the same" if:
            - The table have the same name and the same last modifcation datetime
            - The file have the same name, the same lastm mdoficiation datetime and the same size

        :param other: an other Toolwrapper which maybe as the same inputs
        :type other: :class:`~.wopmars.framework.database.tables.ToolWrapper.ToolWrapper`

        :return: bool
        """
        for t in [t for t in self.tables if t.type.name == "input"]:
            is_same = False
            for t2 in [t2 for t2 in other.tables if t2.type.name == "input"]:
                # two tables are the same if they have the same model/tablename/modfication date
                if (t.model == t2.model and
                    t.tablename == t2.tablename and
                       t.used_at == t2.used_at):
                    is_same = True
                    break
            if not is_same:
                return False

        for f in [f for f in self.files if f.type.name == "input"]:
            is_same = False
            for f2 in [f2 for f2 in other.files if f2.type.name == "input"]:
                # two files are the same if they have the same name, path, size and modficiation date
                if (f.name == f2.name and
                        f.path == f2.path and
                        f.used_at == f2.used_at and
                        f.size == f2.size):
                    is_same = True
                    break
            if not is_same:
                return False
        return True

    def is_output_more_recent_than_input(self):
        """
        Check for files and tables if the outputs are more recent than inputs.

        In a conventionnal use of WoPMaRS, the output are supposed to be younger than the inputs. If they are not,
        we can consider that the input has changed since the last execution and the output has to be re-written.

        :return: Bool: True if the output is actually more recent than input
        """
        most_recent_input = max([datetime.datetime.fromtimestamp(os.path.getmtime(f.path)) for f in self.files if f.type.name == "input"] +
                                [t.modification.date for t in self.tables if t.type.name == "input"])
        oldest_output = min([datetime.datetime.fromtimestamp(os.path.getmtime(f.path)) for f in self.files if f.type.name == "output"] +
                            [t.modification.date for t in self.tables if t.type.name == "output"])
        # in seconds since the begining of time (computer), the oldest thing has a lower number of seconds
        return most_recent_input < oldest_output

    def same_output_than(self, other):
        """
        Never used.

        Check if the output of self is the same than other.

        Checks only if the file names, table names and models are the same.

        :return: bool
        """
        for t in [t for t in self.tables if t.type.name == "output"]:
            is_same = False
            for t2 in [t2 for t2 in other.tables if t2.type.name == "output"]:
                if t.model == t2.model and t.tablename == t2.tablename:
                    is_same = True
                    break
            if not is_same:
                return False

        for f in [f for f in self.files if f.type.name == "input"]:
            is_same = False
            for f2 in [f2 for f2 in other.files if f2.type.name == "input"]:
                if (f.name == f2.name and
                        f.path == f2.path):
                    is_same = True
                    break
            if not is_same:
                return False
        return True

    def does_output_exist(self):
        """
        Check if the output of the ToolWrapper exists.

        For files, it means that the file exists on the system.
        For tables, it means that the table is not empty.

        :return: Bool: True if outputs exist.
        """
        for of in [f for f in self.files if f.type.name == "output"]:
            if not os.path.exists(of.path):
                return False

        for ot in [t for t in self.tables if t.type.name == "output"]:
            if not SQLManager.instance().get_session().query(ot.get_table()).count():
                return False
        return True

    def get_state(self):
        return self.__state

    def set_execution_infos(self, start=None, stop=None, status=None):
        """
        Generic method to set the informations relatives to the execution of the ToolWrapper.

        :param start: The date of start of the Toolwrapper
        :param stop: The date of end of the Toolwrapper
        :param status: The status of the Toolwrapper
        """
        if start is not None:
            self.started_at = start
        if stop is not None:
            self.finished_at = stop
        if self.started_at is not None and self.finished_at is not None:
            self.time = (self.finished_at - self.started_at).total_seconds()
        if status is not None:
            self.status = status

    def set_session(self, session):
        self.__session = session

    def __eq__(self, other):
        """
        Two ToolWrapper objects are equals if all their attributes are equals.

        We check if the files, tables and options are the same.
        :param other: ToolWrapper
        :type other: ToolWrapper
        :return: Bool: True if the ToolWrappers are equals.
        """
        return (isinstance(other, self.__class__) and
                self.same_files(other, "input") and
                self.same_tables(other, "input") and
                self.same_files(other, "output") and
                self.same_tables(other, "output") and
                self.same_options(other))

    def same_files(self, other, type_name):
        """
        Check if the files of a ToolWrapper are the same than the files of the other for a given type (input or output).

        :param other: ToolWrapper with which you need to compare
        :type other: ToolWrapper
        :param type_name: The name of the type of file (input or output)
        :type type_name: str
        :return: Bool: True if the files are the same
        """
        for f in [rf for rf in self.files if rf.type.name == type_name]:
            is_in = bool([rf for rf in other.files if (os.path.abspath(f.path) == os.path.abspath(rf.path) and
                                                       f.name == rf.name and
                                                       rf.type.name == type_name)])
            if not is_in:
                return False
        return True

    def same_tables(self, other, type_name):
        """
        Check if the tables of a ToolWrapper are the same than the tables of the other for a given type (input or output).

        :param other: ToolWrapper with which you need to compare
        :type other: ToolWrapper
        :param type_name: The name of the type of table (input or output)
        :type type_name: str
        :return: Bool: True if the tables are the same
        """
        for t in [t for t in self.tables if t.type.name == type_name]:
            is_in = bool([t for t in other.tables if (t.model == t.model and
                                                      t.type.name == type_name and
                                                      t.tablename == t.tablename)])
            if not is_in:
                return False
        return True

    def same_options(self, other):
        """
        Check if the options of a ToolWrapper are the same the options of the other.

        :param other: ToolWrapper with which you need to compare.
        :type other: ToolWrapper
        :return: Bool: True if the options are the same.
        """
        for opt in self.options:
            is_in = bool([o for o in other.options if (o.name == opt.name and
                                                       o.value == opt.value)])

            if not is_in:
                return False
        return True

    def __hash__(self):
        """
        Redefining the hash method allows ToolWrapper objects to be indexed in sets and dict.

        Needed to use ToolWrapper as nodes of the DiGraph.

        :return: int
        """
        return id(self)

    def __repr__(self):
        """
        Return the string representing the toolwrapper in the DAG.

        :return: String representing the toolwrapper
        """
        s = "\""
        s += "Rule " + self.name
        s += "\\n"
        s += "tool: " + self.__class__.__name__
        s += "\\n"
        for input_f in [f for f in self.files if f.type.name == "input"]:
            s += "\\n\t\t" + input_f.name + ": " + str(input_f.path)
        for input_t in [t for t in self.tables if t.type.name == "input"]:
            s += "\\n\t\tinput_table: " + input_t.name
        s += "\\n"
        for output_f in [f for f in self.files if f.type.name == "output"]:
            s += "\\n\t\t" + output_f.name + ": " + str(output_f.path)
        for output_t in [t for t in self.tables if t.type.name == "output"]:
            s += "\\n\t\toutput_table: " + output_t.name
        s += "\""
        return s

    def dot_label(self):
        """Label for the dot dag"""
        inputs_list_str = [str(i).replace(":", "") for i in self.files + self.tables if i.type.name == "input"]
        outputs_list_str = [str(o).replace(":", "") for o in self.files + self.tables if o.type.name == "output"]
        params_list_str = [str(p).replace(":","") for p in self.options]
        s = ""
        s += "Rule " + self.name + "\n"
        s += "ToolWrapper " + self.__class__.__name__ + "\n"
        s += "Inputs\n" + "\n\t".join(inputs_list_str) + "\n"
        s += "Outputs\n" + "\n".join(outputs_list_str) + "\n"
        s += "Parameters\n" + "\n".join(params_list_str) + "\n"
        return(s)

    def __str__(self):
        inputs_list_str = [str(i) for i in self.files + self.tables if i.type.name == "input"]
        outputs_list_str = [str(o) for o in self.files + self.tables if o.type.name == "output"]
        params_list_str = [str(p) for p in self.options]
        s = ""
        s += "Rule " + str(self.name) + ":" + "\n"
        s += "\ttool: " + str(self.toolwrapper) + "\n"
        if len(inputs_list_str) > 0:
            s += "\tinput:" + "\n"
            s += "\t\t" + "\n\t\t".join(inputs_list_str)
            s += "\n"
        if len(outputs_list_str) > 0:
            s += "\toutput:" + "\n"
            s += "\t\t" + "\n\t\t".join(outputs_list_str)
            s += "\n"
        if len(params_list_str) > 0:
            s += "\tparams:" + "\n"
            s += "\t\t" + "\n\t\t".join(params_list_str)
        return s

    # ###### Method that worker developper should implement#######

    def specify_input_file(self):
        """
        Should be implemented by the toolwrapper developper.

        This method return a List of string containing the input file variable names as String.
        :return: [String]
        """
        return []

    def specify_input_table(self):
        """
        Should be implemented by the toolwrapper developper.

        This method return a List of string containing the input table names names as String.
        :return: [String]
        """
        return []

    def specify_output_file(self):
        """
        Should be implemented by the toolwrapper developper.

        This method return a List of string containing the output file variable names as String.
        :return: [String]
        """
        return []

    def specify_output_table(self):
        """
        Should be implemented by the toolwrapper developper.

        This method return a List of string containing the output table names as String.
        :return: [String]
        """
        return []

    def specify_params(self):
        """
        Should be implemented by the toolwrapper developper.

        This method return a dict of string associated with string. Keys are the name of the options and values, their types.

        :return: {String: String}
        """
        return {}

    def run(self):
        """
        Should be implemented by the toolwrapper developper.

        The core function of the ToolWrapper is this method. It wraps the actual execution of the tool underlying the ToolWrapper.

        :raises NotImplementedError: If it doesn't have been implemented by the ToolWrapper Developer.
        """
        raise NotImplementedError("The method run of the ToolWrapper " + str(self.toolwrapper) + " should be implemented")

    ### Methods availables for the tool developer

    def input_file(self, key):
        """
        Return the path of the specified input file.

        :param key: String the name of the variable containing the path
        :return:
        """
        try:
            return [f.path for f in self.files if f.name == key and f.type.name == "input"][0]
        except IndexError:
            raise WopMarsException("Error during the execution of the ToolWrapper " + str(self.toolwrapper) +
                                   " (rule " + self.name + ").",
                                   "The input file " + str(key) + " has not been specified.")

    def input_table(self, key):
        """
        Return the input table object of the given name.

        :param key: String: the name of the Table object.
        :return:
        """
        try:
            return [t for t in self.tables if t.tablename == key and t.type.name == "input"][0].get_table()
        except IndexError:
            raise WopMarsException("Error during the execution of the ToolWrapper " + str(self.toolwrapper) +
                                   " (rule " + self.name + ").",
                                   "The input table " + str(key) + " has not been specified.")

    def output_file(self, key):
        """
        Return the path of the specified output file.

        :param key: String the name of the variable containing the path
        :return:
        """
        try:
            return [f.path for f in self.files if f.name == key and f.type.name == "output"][0]
        except IndexError:
            raise WopMarsException("Error during the execution of the ToolWrapper " + str(self.toolwrapper) +
                                   " (rule " + self.name + ").",
                                   "The output file " + str(key) + " has not been specified.")

    def output_table(self, key):
        """
        Return the output table object of the given name.

        :param key: String: the name of the Table object.
        :return:
        """
        try:
            return [t for t in self.tables if t.tablename == key and t.type.name == "output"][0].get_table()
        except IndexError:
            raise WopMarsException("Error during the execution of the ToolWrapper " + str(self.toolwrapper) +
                                   " (rule " + self.name + ").",
                                   "The output table " + str(key) + " has not been specified.")

    def option(self, key):
        """
        Return the value associated with the key option.

        If no value is associated with key, return None.

        :param key: The name of the option
        :type key: str
        :return:
        """
        try:
            value = [o.value for o in self.options if o.name == key][0]
            list_splitted_carac = self.specify_params()[key].split("|")
            for s_type in list_splitted_carac:
                s_formated_type = s_type.strip().lower()
                # check if the carac is a castable type
                if s_formated_type in Option.static_option_castable:
                    value = eval(s_formated_type)(value)
                    break
            return value
        except IndexError as e:
            # antipattern, this is bad, but I deleted the warning because if the ToolWrapper Developer put his call to
            # option in a loop, there will be too mutch output
            pass
            return None

    def session(self):
        return self.__session

    def log(self, level, msg):
        """
        use by the toolwrapper developer in order to have a dedicated logger.

        :param level: The level of logging you need: "debug", "info", "warning", "error"
        :type level: str
        :param msg: The actual string to log.
        :type msg: str
        """
        if level == "debug":
            Logger.instance().toolwrapper_debug(msg, self.toolwrapper)
        elif level == "info":
            Logger.instance().toolwrapper_info(msg, self.toolwrapper)
        elif level == "warning":
            Logger.instance().toolwrapper_debug(msg, self.toolwrapper)
        elif level == "error":
            Logger.instance().toolwrapper_error(msg, self.toolwrapper)
        else:
            raise WopMarsException("Error in the Toolwrapper definition of method run()",
                                   "The is no logging level associated with " + str(level) + ". " +
                                   "The authorized ones are: debug, info, warning, error")



import datetime
import time
import unittest
from unittest import TestCase

from test.resource.model.FooBase import FooBase
from test.resource.wrapper.FooWrapper5 import FooWrapper5
from test.resource.wrapper.sprintFive.Add import Add as tw_add
from test.resource.wrapper.sprintFive.Query import Query as tw_query
from wopmars.framework.database.SQLManager import SQLManager
from wopmars.framework.database.tables.IODbPut import IODbPut
from wopmars.framework.database.tables.IOFilePut import IOFilePut
from wopmars.framework.database.tables.ModificationTable import ModificationTable
from wopmars.framework.database.tables.Option import Option
from wopmars.framework.database.tables.Type import Type
from wopmars.framework.management.ToolThread import ToolThread
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.PathFinder import PathFinder


class TestToolThread(TestCase):

    def setUp(self):
        OptionManager.initial_test_setup()
        SQLManager.instance().create_all()

        [SQLManager.instance().get_session().add(FooBase(name="foo " + str(i))) for i in range(10000)]
        SQLManager.instance().get_session().commit()

    def test_run(self):

        input_entry = Type(name="input")
        output_entry = Type(name="output")

        f1 = IOFilePut(name="input1", path="test/resource/input_files/input_file1.txt")
        f1.type = input_entry

        f2 = IOFilePut(name="output1", path="test/output/output_file1.txt")
        f2.type = output_entry

        t1 = IODbPut(model="FooBase", tablename="FooBase")
        t1.set_table(FooBase)
        t1.type = output_entry
        modification_table_entry = ModificationTable(date=datetime.datetime.fromtimestamp(time.time()), table_name=t1.tablename)
        t1.modification = modification_table_entry

        tw1 = FooWrapper5(rule_name="rule1")
        tw1.files.extend([f1, f2])
        tw1.tables.append(t1)

        f12 = IOFilePut(name="input1", path="test/resource/input_files/input_file1.txt")
        f12.type = input_entry

        f22 = IOFilePut(name="output1", path="test/output/output_file1.txt")
        f22.type = output_entry

        t12 = IODbPut(model="FooBase", tablename="FooBase")
        t12.set_table(FooBase)
        t12.type = output_entry
        modification_table_entry = ModificationTable(
            date=datetime.datetime.fromtimestamp(time.time()), table_name=t12.tablename)
        t12.modification = modification_table_entry

        tw2 = FooWrapper5(rule_name="rule2")
        tw2.files.extend([f12, f22])
        tw2.tables.append(t12)

        f13 = IOFilePut(name="input1", path="test/resource/input_files/input_file1.txt")
        f13.type = input_entry

        f23 = IOFilePut(name="output1", path="test/output/output_file1.txt")
        f23.type = output_entry

        t13 = IODbPut(model="FooBase", tablename="FooBase")
        t13.set_table(FooBase)
        t13.type = output_entry
        modification_table_entry = ModificationTable(
            date=datetime.datetime.fromtimestamp(time.time()), table_name=t13.tablename)
        t13.modification = modification_table_entry

        tw3 = FooWrapper5(rule_name="rule3")
        tw3.files.extend([f13, f23])
        tw3.tables.append(t13)

        tt1 = ToolThread(tw1)
        tt2 = ToolThread(tw2)
        tt3 = ToolThread(tw3)

        tt1.start()
        tt2.start()
        tt3.start()

        tt1.join()
        tt2.join()
        tt3.join()

        self.assertEqual(len(SQLManager.instance().get_session().query(FooBase).filter(FooBase.name.like('Foowrapper5 - %')).all()), 3000)

    def test_run_commit_vs_query(self):
        input_entry = Type(name="input")
        output_entry = Type(name="output")

        f1 = IOFilePut(name="input1", path="test/resource/input_files/input_file1.txt")
        f1.type = input_entry

        t1 = IODbPut(model="FooBase", tablename="FooBase")
        t1.set_table(FooBase)
        t1.type = output_entry
        modification_table_entry = ModificationTable(date=datetime.datetime.fromtimestamp(time.time()), table_name=t1.tablename)
        t1.modification = modification_table_entry

        o1 = Option(name="rows", value="1000")

        tw1 = tw_add(rule_name="rule1")
        tw1.files.append(f1)
        tw1.tables.append(t1)
        tw1.options.append(o1)

        f12 = IOFilePut(name="input1", path="test/resource/input_files/input_file1.txt")
        f12.type = input_entry

        t12 = IODbPut(model="FooBase", tablename="FooBase")
        t12.set_table(FooBase)
        t12.type = output_entry
        modification_table_entry = ModificationTable(date=datetime.datetime.fromtimestamp(time.time()),
                                                     table_name=t12.tablename)
        t12.modification = modification_table_entry

        o12 = Option(name="rows", value="1000")

        tw12 = tw_add(rule_name="rule1")
        tw12.files.append(f12)
        tw12.tables.append(t12)
        tw12.options.append(o12)

        f13 = IOFilePut(name="input1", path="test/resource/input_files/input_file1.txt")
        f13.type = input_entry

        t13 = IODbPut(model="FooBase", tablename="FooBase")
        t13.set_table(FooBase)
        t13.type = output_entry
        modification_table_entry = ModificationTable(date=datetime.datetime.fromtimestamp(time.time()),
                                                     table_name=t13.tablename)
        t13.modification = modification_table_entry

        o13 = Option(name="rows", value="1000")

        tw13 = tw_add(rule_name="rule1")
        tw13.files.append(f13)
        tw13.tables.append(t13)
        tw13.options.append(o13)

        tt1 = ToolThread(tw1)
        tt2 = ToolThread(tw12)
        tt3 = ToolThread(tw13)

        t21 = IODbPut(model="FooBase", tablename="FooBase")
        t21.set_table(FooBase)
        t21.type = input_entry

        tw21 = tw_query(rule_name="rule1")
        tw21.tables.append(t21)

        t22 = IODbPut(model="FooBase", tablename="FooBase")
        t22.set_table(FooBase)
        t22.type = input_entry

        tw22 = tw_query(rule_name="rule1")
        tw22.tables.append(t22)

        t23 = IODbPut(model="FooBase", tablename="FooBase")
        t23.set_table(FooBase)
        t23.type = input_entry

        tw23 = tw_query(rule_name="rule1")
        tw23.tables.append(t23)

        tt4 = ToolThread(tw21)
        tt5 = ToolThread(tw22)
        tt6 = ToolThread(tw23)

        tt4.start()

        tt1.start()
        tt2.start()
        tt3.start()
        time.sleep(5)
        tt5.start()
        tt6.start()

        tt1.join()
        tt2.join()
        tt3.join()
        tt4.join()
        tt5.join()
        tt6.join()

    def tearDown(self):
        SQLManager.instance().get_session().close()
        SQLManager.instance().drop_all()
        PathFinder.silentremove("test/output/")

if __name__ == '__main__':
    unittest.main()



from wopmars.framework.database.tables.ToolWrapper import ToolWrapper

# import models
import wopexamplesnp.model.SNP

import csv

class InsertSnp(ToolWrapper): # inherit WopMars
    __mapper_args__ = {'polymorphic_identity': __name__}
    __input_file_snp = "snp"
    __output_table_snp = "SNP"

    def specify_input_file(self):
        return [InsertSnp.__input_file_snp]

    def specify_output_table(self):
        return [InsertSnp.__output_table_snp]

    def run(self):
        session = self.session()
        #
        snp_path = self.input_file(InsertSnp.__input_file_snp)
        snp_model = self.output_table(InsertSnp.__output_table_snp)
        #
        # read input file
        input_file_dic_list = []
        for line in csv.reader(open(snp_path, 'r', encoding='utf-8'), delimiter="\t"):
            chrom = line[0]
            position = line[1]
            rsid = line[2]
            input_file_obj = {'chrom': chrom, 'position': position, 'rsid': rsid}
            try: # checks if exists SNP in db
                session.query(snp_model).filter_by(**input_file_obj).one()
            except: # if not, add
                session.add(snp_model(**input_file_obj))
            session.commit()





import queue
import unittest
from unittest import TestCase

from wopmars.utils.UniqueQueue import UniqueQueue


class TestUniqueQueue(TestCase):
    def setUp(self):
        self.__queue = UniqueQueue()
        self.__queue.put(1)
        self.__queue.put(2)
        self.__queue.put(1)

    def test_put(self):
        try:
            self.__queue.get_nowait()
            self.__queue.get_nowait()
        except:
            AssertionError("Should not raise an exception")

        self.assertRaises(queue.Empty, self.__queue.get_nowait)

    def test_get_queue_tuple(self):
        self.assertEqual(len(self.__queue.get_queue_tuple()), 2)

if __name__ == '__main__':
    unittest.main()


import os
import unittest
from unittest import TestCase

from test.resource.wrapper.FooWrapper10 import FooWrapper10
from test.resource.wrapper.FooWrapper4 import FooWrapper4
from test.resource.wrapper.FooWrapper5 import FooWrapper5
from test.resource.wrapper.FooWrapper6 import FooWrapper6
from test.resource.wrapper.FooWrapper7 import FooWrapper7
from test.resource.wrapper.FooWrapper8 import FooWrapper8
from test.resource.wrapper.FooWrapper9 import FooWrapper9
from wopmars.framework.database.SQLManager import SQLManager
from wopmars.framework.database.tables.IODbPut import IODbPut
from wopmars.framework.database.tables.IOFilePut import IOFilePut
from wopmars.framework.database.tables.Type import Type
from wopmars.framework.management.DAG import DAG
from wopmars.framework.parsing.Parser import Parser
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.PathFinder import PathFinder
from wopmars.utils.exceptions.WopMarsException import WopMarsException


class TestParser(TestCase):
    def setUp(self):
        OptionManager.initial_test_setup()

        SQLManager.instance().create_all()
        session = SQLManager.instance().get_session()
        session.get_or_create(Type, defaults={"id": 1}, name="input")
        session.get_or_create(Type, defaults={"id": 2}, name="output")
        session.commit()
        self.__s_root_path = PathFinder.get_module_path()
        self.__parser = Parser()

    def tearDown(self):
        SQLManager.instance().get_session().close()
        SQLManager.instance().drop_all()
        PathFinder.dir_content_remove("test/output")
        OptionManager._drop()
        SQLManager._drop()

    def test_parse(self):
        OptionManager.initial_test_setup()

        # The good --------------------------:
        input_entry = Type(name="input")
        output_entry = Type(name="output")

        f1 = IOFilePut(name="input1", path="test/resource/input_files/input_file1.txt")
        f1.type = input_entry

        f2 = IOFilePut(name="output1", path="test/output/output_file1.txt")
        f2.type = output_entry

        f3 = IOFilePut(name="input1", path="test/output/output_file1.txt")
        f3.type = input_entry

        f3bis = IOFilePut(name="input1", path="test/output/output_file1.txt")
        f3bis.type = input_entry

        f4 = IOFilePut(name="output1", path="test/output/output_file2.txt")
        f4.type = output_entry

        f5 = IOFilePut(name="output1", path="test/output/output_file3.txt")
        f5.type = output_entry

        f6 = IOFilePut(name="output2", path="test/output/output_file4.txt")
        f6.type = output_entry

        f7 = IOFilePut(name="input1", path="test/output/output_file3.txt")
        f7.type = input_entry

        f8 = IOFilePut(name="input2", path="test/output/output_file2.txt")
        f8.type = input_entry

        f9 = IOFilePut(name="output1", path="test/output/output_file5.txt")
        f9.type = output_entry

        f10 = IOFilePut(name="input1", path="test/output/output_file4.txt")
        f10.type = input_entry

        f11 = IOFilePut(name="output1", path="test/output/output_file6.txt")
        f11.type = output_entry

        f12 = IOFilePut(name="input1", path="test/output/output_file1.txt")
        f12.type = input_entry

        f13 = IOFilePut(name="input2", path="test/output/output_file5.txt")
        f13.type = input_entry

        f14 = IOFilePut(name="input3", path="test/output/output_file6.txt")
        f14.type = input_entry

        f15 = IOFilePut(name="output1", path="test/output/output_file7.txt")
        f15.type = output_entry

        t1 = IODbPut(model="FooBase", tablename="FooBase")
        t1.type = output_entry

        t1bis = IODbPut(model="FooBase", tablename="FooBase")
        t1bis.type = input_entry

        t2 = IODbPut(model="FooBase2", tablename="FooBase2")
        t2.type = output_entry

        t2bis = IODbPut(model="FooBase2", tablename="FooBase2")
        t2bis.type = input_entry

        tw1 = FooWrapper4(rule_name="rule1")
        tw1.files.extend([f1, f2])
        tw2 = FooWrapper5(rule_name="rule2")
        tw2.files.extend([f3, f4])
        tw2.tables.extend([t1])
        tw3 = FooWrapper6(rule_name="rule3")
        tw3.files.extend([f3bis, f5, f6])
        tw4 = FooWrapper7(rule_name="rule4")
        tw4.tables.extend([t1bis, t2])
        tw5 = FooWrapper8(rule_name="rule5")
        tw5.files.extend([f8, f7, f9])
        tw6 = FooWrapper9(rule_name="rule6")
        tw6.files.extend([f10, f11])
        tw6.tables.extend([t2bis])
        tw7 = FooWrapper10(rule_name="rule7")
        tw7.files.extend([f12, f13, f14, f15])

        set_toolwrappers = set([tw1, tw2, tw3, tw4, tw5, tw6, tw7])

        OptionManager.instance()["--dot"] = None

        dag_expected = DAG(set_toolwrappers)
        OptionManager.instance()["--wopfile"] = os.path.join(self.__s_root_path, "test/resource/wopfile/example_def_file1.yml")
        dag_obtained = self.__parser.parse()
        self.assertEqual(dag_expected, dag_obtained)

        OptionManager.instance()["--wopfile"] = os.path.join(self.__s_root_path, "test/resource/wopfile/example_def_file_not_a_dag.yml")
        with self.assertRaises(WopMarsException):
            self.__parser.parse()

        # Verify the dot file ----------------:
        OptionManager.instance()["--wopfile"] = os.path.join(self.__s_root_path, "test/resource/wopfile/example_def_file1.yml")
        #dot_path = os.path.join(self.__s_root_path, "test_bak.dot")
        #OptionManager.instance()["--dot"] = dot_path
        self.__parser.parse()
        #self.assertTrue(os.path.isfile(dot_path))
        #os.remove(dot_path)
        #os.remove(dot_path[:-4] + ".ps")

if __name__ == '__main__':
    unittest.main()



"""
Module containing the FooWrapper1 class
"""
import time

from test.resource.model.FooBase2 import FooBase2
from wopmars.framework.database.tables.ToolWrapper import ToolWrapper

class FooWrapper7(ToolWrapper):
    """
    This class has been done for example/testing purpose.
    Modifications may lead to failure in tests.
    """
    __mapper_args__ = {'polymorphic_identity': "FooWrapper7"}

    def specify_input_table(self):
        return ["FooBase"]

    def specify_output_table(self):
        return ["FooBase2"]

    def run(self):
        print(self.__class__.__name__ + " is running...")
        inputs = self.session().query(self.input_table("FooBase")).all()
        self.session().delete_content(self.output_table("FooBase2"))
        for i in inputs:
            entry = self.output_table("FooBase2")(name=i.name)
            self.session().add(entry)
        print(self.session().query(FooBase2).all())
        time.sleep(1)



import threading
import unittest
from unittest import TestCase

from test.resource.model.FooBase import FooBase
from wopmars.framework.database.SQLManager import SQLManager
from wopmars.utils.OptionManager import OptionManager


class ConcurrentCommitingThread(threading.Thread):
    def run(self):
        thread_session = SQLManager.instance().get_session()
        for i in range(1000):
            foo = FooBase(name="string " + str(i))
            thread_session.add(foo)
        thread_session.commit()

        thread_session.close()


# todo LG: refaire test_bak rollback
class ConcurrentRollBackingThread(threading.Thread):
    def run(self):
        thread_session = SQLManager.instance().get_session()
        for i in range(1000):
            foo = FooBase(name="string " + str(i))
            thread_session.add(foo)
        thread_session.rollback()

        thread_session.close()


class TestSQLManager(TestCase):
    def setUp(self):
        OptionManager.initial_test_setup()
        SQLManager.instance().create_all()
        self.__local_session = SQLManager.instance().get_session()

        SQLManager()
        self.__t1 = ConcurrentCommitingThread()
        self.__t2 = ConcurrentCommitingThread()
        self.__t3 = ConcurrentCommitingThread()

        self.__t4 = ConcurrentRollBackingThread()
        self.__t5 = ConcurrentRollBackingThread()
        self.__t6 = ConcurrentRollBackingThread()

    def test_commit(self):
        list_threads = [self.__t1, self.__t2, self.__t3, self.__t4, self.__t5, self.__t6]
        try:
            for t in list_threads:
                t.start()

            for t in list_threads:
                t.join()
        except Exception as e:
            print(e)
            raise AssertionError("Should not raise an exception")

        self.assertTrue(len(self.__local_session.query(FooBase).filter(FooBase.name.like('string %')).all()) == 3000)

    def tearDown(self):
        SQLManager.instance().get_session().close()
        SQLManager.instance().drop_all()
        OptionManager._drop()
        SQLManager._drop()

if __name__ == '__main__':
    unittest.main()



import subprocess

import networkx as nx
from networkx.drawing.nx_pydot import write_dot

from wopmars.framework.database.SQLManager import SQLManager
from wopmars.utils.Logger import Logger
from wopmars.utils.SetUtils import SetUtils


class DAG(nx.DiGraph):
    """
    This class inherits from networkx.DiGraph class and is able to represent a DAG of tool nodes. It takes a set of
    :class:`~.wopmars.framework.database.tables.ToolWrapper` and analyse it to extract dependencies between them.
    """    
    def __init__(self, set_tools=None):
        """
        The DAG can be build from a set of tools, analyzing the successors of each of them.

        ToolWrappers has a method "follows()" wich allow to know if one tool has a dependency for one other. The tools
        of the set_tools are compared each other to extract the dependencies.
        
        :param set_tools: A set of tools
        """
        # the DAG is a DiGraph
        super().__init__()
        # A nx digraph to store the dot graph
        self.dot_digraph = nx.DiGraph()
        Logger.instance().info("Building the execution DAG...")
        if set_tools:
            # for each tool
            for tool1 in set_tools:
                self.add_node(tool1)
                # for each other tool
                for tool2 in set_tools.difference(set([tool1])):
                    # is there a dependency between tool1 and tool2?
                    if tool1.follows(tool2):
                        self.add_edge(tool2, tool1)
        if set_tools:
            # for each tool
            for tool1 in set_tools:
                self.dot_digraph.add_node(tool1.dot_label())
                # for each other tool
                for tool2 in set_tools.difference(set([tool1])):
                    # is there a dependency between tool1 and tool2?
                    if tool1.follows(tool2):
                        self.dot_digraph.add_edge(tool2.dot_label(), tool1.dot_label())
        Logger.instance().debug("DAG built.")

    def write_dot(self, path):
        """
        Build the dot file.

        The .ps can be built from the dot file with the command line: "dot -Tps {filename}.dot - o {filename}.ps"
        """
        # To build .ps : dot -Tps {filename}.dot - o {filename}.ps
        nx.draw(self.dot_digraph)
        write_dot(self.dot_digraph, path)
        # building the openable file:
        list_popen = ["dot", "-Tps", path, "-o", path.rsplit("/", 1)[0] + "/" + path.rsplit("/", 1)[1].split(".")[-2] + ".ps"]
        Logger.instance().debug("SubProcess command line for .ps file: " + str(list_popen))
        p = subprocess.Popen(list_popen)
        p.wait()

    def successors(self, node):
        """
        Get the successors of a node.

        The method is overwhelmed because if a node is None, then, the root nodes are returned.

        :param node: a node of the DAG or None.
        :type node: :class:`~.wopmars.framework.database.tables.ToolWrapper.ToolWrapper`
        :return: [node]:  the successors of the given node or the node at the root of the DAG.
        """
        if node is None:
            # in_degree is the number of incoming edges to a node. If the degree is 0, then the node is at the root
            # of the DAG.
            # return [n for n, d in self.in_degree().items() if d == 0]
            return [n for n, d in list(self.in_degree()) if d == 0]
        else:
            s = list(super().successors(node))
            return s

    def get_all_successors(self, node):
        """
        Return the set of all successors nodes from a given node in the DAG (node included).

        :param node: a node of the DAG or None.
        :type node: :class:`~.wopmars.framework.database.tables.ToolWrapper.ToolWrapper`

        :return: set(node): all the successors of a given node.
        """
        list_successors = [node]
        for N in self.successors(node):
            list_successors.extend(self.get_all_successors(N))
        return set(list_successors)

    def get_all_predecessors(self, node):
        """
        Return the set of all predecessors nodes from a given node in the DAG (node included).

        :param node: a node of the DAG or None.
        :type node: :class:`~.wopmars.framework.database.tables.ToolWrapper.ToolWrapper`

        :return: set(node): all the predecessors of a given node.
        """
        list_predecessors = [node]
        for N in self.predecessors(node):
            list_predecessors.extend(self.get_all_predecessors(N))
        return set(list_predecessors)

    def __eq__(self, other):
        """
        Test if self equals other.

        Check the number of nodes and the set of edges of each graphs.

        :param other: the other DAG to compare
        :type other: :class:`wopmars.framework.management.DAG.DAG`

        :return: True if self == other. Else, False.
        """
        assert isinstance(other, self.__class__)
        int_nodes_self = len(self.nodes())
        int_nodes_other = len(other.nodes())

        set_edges_self = set(self.edges())
        set_edges_other = set(other.edges())

        return (
            int_nodes_self == int_nodes_other and
            SetUtils.all_elm_of_one_set_in_one_other(set_edges_self, set_edges_other) and
            SetUtils.all_elm_of_one_set_in_one_other(set_edges_other, set_edges_self)
        )



"""
Module containing the UniqueQueue class.
"""
from queue import Queue


class UniqueQueue(Queue):
    """
    The UniqueQueue re-implement Queue, modificating its _put method to ensure that each element is unique.
    """

    def _put(self, item):
        """
        Check if item is already in queue. If not, append the queue and return True, else, return False.

        :param item: an item that has to be added to the Queue
        :return: bool for the success of the put method
        """
        if item not in self.queue:
            self.queue.append(item)
            return True
        return False

    def get_queue_tuple(self):
        """
        Return the ordered content of the queue in a tuple.

        :return: tuple of the elements in the Queue.
        """
        return tuple(self.queue)



from wopmars.framework.database.Base import Base

from sqlalchemy import Column, Integer, String, Float


class Piece(Base):
    __tablename__ = "piece"

    id = Column(Integer, primary_key=True, autoincrement=True)
    serial_number = Column(String, unique=True)
    type = Column(String)
    price = Column(Float)


from setuptools import setup, find_packages

setup(
    name='wopexamplecar',
    version='1.0',
    description='Simple car example for experimenting Wopmars',
    author='Luc Giffon - TAGC',
    author_email='luc.giffon@gmail.com',
    packages=find_packages(),
)



"""
Example of module documentation which can be
multiple-lined
"""
from sqlalchemy import Column, Integer, String

from wopmars.framework.database.Base import Base


class FooBase2P(Base):
    """
    Documentation for the class
    """
    __tablename__ = "FooBase2P"

    id = Column(Integer, primary_key=True)
    name = Column(String(255))


"""
Module containing the SQLManager class.
"""
# import pandas
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy import create_engine
from sqlalchemy.schema import sort_tables
from sqlalchemy import event

from wopmars.framework.database.Base import Base
from wopmars.framework.database.WopMarsSession import WopMarsSession
from wopmars.utils.Logger import Logger
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.RWLock import RWLock
from wopmars.utils.Singleton import SingletonMixin
from wopmars.utils.exceptions.WopMarsException import WopMarsException


class SQLManager(SingletonMixin):
    """
    The class SQLManager is responsible of all operations performed on the database.

    The SQLManager is a synchronized Singleton which may be accessible from any thread, at any time. This is allowed
    by the SingletonMixin superclass and a read/write lock.

    Provides a WopMarsSession through the method "get_session". The session is used to perform calls to the database.
    Provides the engine if necessary through "get_engine".
    Allows commiting and rollbacking operations of a given session through methods "commit" and "rollback".
    Allows querying operations from a given query: all, one, first, count, one_or_none, scalar. Through "result_factory".
    Allows executing statements through the "execute" method.
    Allows droping and creating tables from table names ("create" and "drop"), list of table models ("drop_table_list")
    and all ("create_all" and "drop_all").
    """
    wom_table_names = [
        "wom_type",
        "wom_modification_table",
        "wom_execution",
        "wom_rule",
        "wom_table",
        "wom_file",
        "wom_option"
    ]

    def __init__(self):
        """
        The constructor of the SQLManager is supposed to be called once in the whole execution, thanks to the
        SingletonMixin inheritance.

        The constructor create the engine at the localization provided by the user. The option "connect_args" is set to
        False.

        The "PRAGMA foreign_keys=ON" statement is executed here and allows to enforce foreign_keys constraints.

        The Session attribute is a object of class "Type" and allows to make a scoped session bound to the engine on demand.
        The autocommit is set to False and the autoflush, to True.

        The lock is initialized here thanks to the RW lock class and will be used to overide the behaviour of sqlite
        to assess the access to the databse without using the queue of SQLite, bound at 4 sec wait before error.
        """
        s_database_url = OptionManager.instance()["--database"]
        self.d_database_config = {'db_connection': None, 'db_database': None, 'db_url': s_database_url}
        self.d_database_config['db_connection'] = s_database_url.split("://")[0]
        if self.d_database_config['db_connection'] == "sqlite":
            self.d_database_config['db_database'] = s_database_url.replace("sqlite:///", "")
        # echo=False mute the log of database
        # connect_args have been necessary because of the accession of the same objects in different Threads.
        if self.d_database_config['db_connection']=="sqlite":
            self.__engine = create_engine(self.d_database_config['db_url'], echo=False, connect_args={'check_same_thread': False})
        else:
            self.__engine = create_engine(self.d_database_config['db_url'], echo=False)
        # Below, between "###", code copy-pasted from this post
        # http://stackoverflow.com/questions/2614984/sqlite-sqlalchemy-how-to-enforce-foreign-keys/7831210#7831210
        # enforce foreign key constraints
        ###
        def _fk_pragma_on_connect(dbapi_con, con_record):
            if self.d_database_config['db_connection'] == "sqlite":
                dbapi_con.execute('pragma foreign_keys=ON')

        event.listen(self.__engine, 'connect', _fk_pragma_on_connect)
        ###

        # I don't know why I have used the autoflush=True
        self.__Session = scoped_session(sessionmaker(bind=self.__engine, autoflush=True, autocommit=False))
        # The lock
        self.__lock = RWLock()

    def get_engine(self):
        """
        Return the engine of the current execution.

        :return: engine
        """
        return self.__engine

    def get_session(self):
        """
        Return a WopmarsSession associated with this SQLmanager.

        :return: WopMarsSession
        """
        session = WopMarsSession(self.__Session(), self)
        return session

    def commit(self, session):
        """
        Commit the current session.

        The SQLManager wrap the sqlite queue for commiting operations on database in order to do not trigger the error due
        to the time-out operations.

        This is done thanks to a read_write_lock: sqlmanager is a synchronized singleton with synchronized methods.

        :param session: SQLAlchemy session object
        """
        try:
            Logger.instance().debug(str(session) + " want the write lock on SQLManager.")
            self.__lock.acquire_write()
            Logger.instance().debug(str(session) + " has taken the write lock on SQLManager.")
            session.commit()
            Logger.instance().debug(str(session) + " has been commited.")
        finally:
            # Always release the lock
            self.__lock.release()
            Logger.instance().debug(str(session) + " has released the write lock on SQLManager.")

    def result_factory(self, query, method):
        """
        Return the result of the query, using the demanded method.

        The result_factory wrap the methods for querying database:
          - all
          - first
          - one
          - one_or_none
          - scalar
          - count

        This is necessary to use the ReadLock of WopMars instead of the one from SQLite.

        :param query: The query object, ready to be performed.
        :param method: String signifying the method which have to be used for querying database
        :return: The result of the query.
        """
        result = None
        try:
            Logger.instance().debug("Executing query on session " + str(query.session) + ": \n" + str(query) + ";")
            Logger.instance().debug("WopMarsQuery " + str(query.session) + " want the read-lock on SQLManager")
            self.__lock.acquire_read()
            Logger.instance().debug("\"" + str(query.session) + "\" has taken the read lock on SQLManager.")
            # switch case according to the demanded method.
            # in each condition, we call the superclass associated method: superclass is Query from sqlalchemy
            if method == "all":
                result = super(query.__class__, query).all()
            elif method == "one":
                result = super(query.__class__, query).one()
            elif method == "first":
                result = super(query.__class__, query).first()
            elif method == "count":
                result = super(query.__class__, query).count()
            elif method == "one_or_none":
                result = super(query.__class__, query).one_or_none()
            elif method == "scalar":
                result = super(query.__class__, query).scalar()
            else:
                raise WopMarsException("Error while querying the database.",
                                       "Demanded operation doesn't exist: " + str(method))
        finally:
            # Always release the lock
            self.__lock.release()
            Logger.instance().debug("\"" + str(query.session) + "\" has released the read lock on SQLManager.")
        return result

    def execute(self, session, statement, *args, **kwargs):
        """
        Allow to execute a statement object on the given session.

        :param session: SQLAlchemy session object.
        :param statement: SQLAlchemy statement object.
        """
        Logger.instance().debug("SQLManager.execute(" + str(session) + ", " + str(statement) + ", " +
                                    str(args) + ", " + str(kwargs) + ")")
        try:
            Logger.instance().debug(str(session) + " want the write lock on SQLManager for statement \"" + str(statement) + "\"")
            self.__lock.acquire_write()
            Logger.instance().debug(str(session) + " has taken the write lock on SQLManager.")
            return session.execute(statement, *args, **kwargs)
        finally:
            # Always release the lock
            self.__lock.release()
            Logger.instance().debug(str(session) + " has released the write lock on SQLManager.")

    def rollback(self, session):
        """
        Rollback the given session.

        The SQLManager wrap the sqlite queue for rollbacking operations on database in order to do not trigger the error due
        to the time-out operations.

        This is done thanks to a read_write_lock: sqlmanager is a synchronized singleton with synchronized methods.

        :param session: SQLAlchemy session object
        """
        try:
            Logger.instance().debug(str(session) + " want the write lock on SQLManager.")
            self.__lock.acquire_write()
            Logger.instance().debug(str(session) + " has taken the write lock on SQLManager.")
            session.rollback()
            Logger.instance().debug(str(session) + " has been rollbacked.")
        finally:
            # Always release the lock
            self.__lock.release()
            Logger.instance().debug(str(session) + " has released the write lock on SQLManager.")

    def drop_all(self):
        """
        Use the declarative Base to drop all tables found. Should only be used for testing purposes.
        """
        try:
            self.__lock.acquire_write()
            Logger.instance().debug("Dropping all tables...")
            Base.metadata.drop_all(self.__engine)
        finally:
            # Always release the lock
            self.__lock.release()

    def create_all(self):
        """
        Use the declarative Base to create all tables found.

        If you want to create a table, you should be sure that it has been imported first.
        """
        try:
            self.__lock.acquire_write()
            Logger.instance().debug("Creating all tables...")
            Base.metadata.create_all(self.__engine)
            # Always release the lock
        finally:
            self.__lock.release()

    def create(self, tablename):
        """
        Use the declarative Base to create a table from its tablename.

        The tablename is the name of the base represented in the databse (independent of the Table model)
        :param tablename:
        """
        try:
            self.__lock.acquire_write()
            Logger.instance().debug("SQLManager.create(" + str(tablename) + "): create table " + str(tablename))
            Base.metadata.tables[tablename.split(".")[-1]].create(self.__engine, checkfirst=True)
        finally:
            # Always release the lock
            self.__lock.release()

    def drop(self, tablename):
        """
        Use the declarative Base to drop a table from its tablename.

        The tablename is the name of the base represented in the databse (independent of the Table model)
        :param tablename:
        """
        try:
            self.__lock.acquire_write()
            # todo tabling
            Logger.instance().debug("SQLManager.drop(" + str(tablename) + "): drop table " + str(tablename))
            Base.metadata.tables[tablename.split(".")[-1]].drop(self.__engine, checkfirst=True)
        finally:
            # Always release the lock
            self.__lock.release()

    def drop_table_list(self, list_str_table):
        """
        Remove a list of tables from the list of their tablenames.

        :param list_str_table: [String] the name of the tables.
        """
        # Get the list of Table objects from tablenames, then sort them according to their relationships / foreignkeys
        # and take the reverse to delete them in the right order (reverse of the correct order for creating them)
        # todo tabling
        list_obj_table = reversed(sort_tables([Base.metadata.tables[tablename.split(".")[-1]] for tablename in list_str_table]))
        try:
            self.__lock.acquire_write()
            for t in list_obj_table:
                Logger.instance().debug("SQLManager.drop_table_list(" + str(list_str_table) + "): drop table " + str(t.name))
                t.drop(self.__engine, checkfirst=True)
        finally:
            # Always release the lock
            self.__lock.release()

    def drop_table_content_list(self, list_str_table):
        """
        Remove a list of tables from the list of their tablenames.

        :param list_str_table: [String] the name of the tables.
        """
        session = self.get_session()
        # Get the list of Table objects from tablenames, then sort them according to their relationships / foreignkeys
        # and take the reverse to delete them in the right order (reverse of the correct order for creating them)
        list_obj_table = reversed(
            sort_tables([Base.metadata.tables[tablename.split(".")[-1]] for tablename in list_str_table]))
        for t in list_obj_table:
            Logger.instance().debug(
                "SQLManager.drop_table_content_list(" + str(list_str_table) + "): drop table content " + str(t.name))
            self.execute(session._session(), t.delete())

    # def pandas_to_sql(self, df, *args, **kwargs):
    #     """
    #     Execute the `DataFrame.to_sql <http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.to_sql.html>`_ function from pandas.DataFrame.
    #
    #     :param df: The DataFrame to insert in database.
    #     :type df: `pandas.DataFrame <http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.html>`_
    #     :param args: The conventional arguments for the pandas.to_sql function.
    #     :param kwargs: The conventional key arguments for the pandas.to_sql function.
    #     """
    #     try:
    #         self.__lock.acquire_write()
    #         df.to_sql(*args, **kwargs)
    #         Logger.instance().debug("SQLManager.pandas_to_sql: Adding dataframe to database")
    #     finally:
    #         self.__lock.release()
    #
    # def pandas_read_sql(self, *args, **kwargs):
    #     """
    #     Execute the `pandas.read_sql <http://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_sql.html?highlight=sql>`_ function
    #
    #     :param args: The conventional arguments for the pandas.read_sql function.
    #     :param kwargs: The conventional key arguments for the pandas.read_sql function.
    #     :return DataFrame: The dataframe containing the results of the query.
    #     """
    #     try:
    #         self.__lock.acquire_read()
    #         df = pandas.read_sql(*args, **kwargs)
    #         Logger.instance().debug("SQLManager.read_sql: Reading database using pandas")
    #     finally:
    #         self.__lock.release()
    #     return df

    def create_trigger(self, table, ddl):
        """
        Create a Trigger after creating a given table.

        :param table:
        :type table: `Table <http://docs.sqlalchemy.org/en/latest/core/metadata.html#sqlalchemy.schema.Table>`_
        :param ddl:
        :type ddl: `DDL <http://docs.sqlalchemy.org/en/latest/core/ddl.html#sqlalchemy.schema.DDL>`_
        """
        try:
            self.__lock.acquire_write()
            event.listen(table, 'after_create', ddl)
        finally:
            self.__lock.release()




import datetime
import sys

import time

from wopmars.framework.database.SQLManager import SQLManager
from wopmars.framework.database.tables.Execution import Execution
from wopmars.framework.database.tables.IODbPut import IODbPut
from wopmars.framework.database.tables.ToolWrapper import ToolWrapper
from wopmars.framework.database.tables.Type import Type
from wopmars.framework.management.DAG import DAG
from wopmars.framework.management.ToolThread import ToolThread
from wopmars.framework.management.ToolWrapperObserver import ToolWrapperObserver
from wopmars.framework.parsing.Parser import Parser
from wopmars.utils.PathFinder import PathFinder
from wopmars.utils.Logger import Logger
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.UniqueQueue import UniqueQueue
from wopmars.utils.exceptions.WopMarsException import WopMarsException


class WorkflowManager(ToolWrapperObserver):
    """
    The WorkflowManager class manage all the workflow execution.

    WorkflowManager will ask the parser to build the DAG then execute it. The execution of the DAG is done as following:

    1- The workflow manager take the "tool_dag" with all the rules of the workflow and build the "dag_to_exec" which
    will be actually executed according to the "target rule" and "source rule" options.
    2- The method :meth:`~.wopmars.framework.management.WorkflowManager.WorkflowManager.execute_from` is call without argument, meaning that the execution begin at the top of the dag
    (the root of the tree).
    3- The nodes are gathered thanks to the :meth:`~.wopmars.framework.management.DAG.DAG.successors` method of the :class:`wopmars.framework.management.DAG.DAG`
    4- Each node is wrapped inside a :class:`~.wopmars.framework.management.ToolThread.ToolThread` object which will be added to the queue.
    5- Each ToolThread is executed (ordered) as follows:

      a- If the inputs are ready: they are executed.
      b- If not, they are put in the buffer.

    6- When the :class:`~.wopmars.framework.management.ToolThread.ToolThread` has finished its execution, a notification of success is sent.
    7- The method :meth:`~wopmars.framework.management.WorkflowManager.WorkflowManager.execute_from` is called again with the succeeded ToolWrapper as argument.
    8- Loop to the 3rd step
    9- When the DAG is finished, the software exits

    Every exception or error are raised to the top level of the software through WopMarsException with
    a String explaining the context and details about the exception.
    """    
    def __init__(self):
        """
        The parser will give the DAG which will be executed.
        The queue_exec is the Thread pool. It will contains the tool threads that will wait for being executed. Each tool
        should appear only once in the queue.
        The list_queue_buffer will be filled with the tool threads that the WorkflowManager couldn't execute.
        The count_exec is a counter that keep trace of the number of tools that are currently executed.
        The dag_tools will contain the dag representing the workflow.
        The dag_to_exec is basically the same dag than dag_tools or a subgraph depending on the options --sourcerule or --targetrule
        given by the user.
        The session is used to get back the session without calling again SQLManager.
        """
        self.__parser = Parser()
        self.__queue_exec = UniqueQueue()
        self.__list_queue_buffer = []
        self.__count_exec = 0
        self.__dag_tools = None
        self.__dag_to_exec = None
        self.__already_runned = set()
        self.__session = SQLManager.instance().get_session()

    def run(self):
        """
        Get the dag then execute it.

        The database is setUp here if workflow side tables have not been created yet.

        The dag is taken thanks to the :meth:`~.wopmars.framework.parsing.Parser.Parser.parse` method of the parser. And then pruned by the :meth:`~.wopmars.framework.management.WorkflowManager.WorkflowManager.get_dag_to_exec` method
        which will set the right DAG to be executed.
        Then, :meth:`~.wopmars.framework.management.WorkflowManager.WorkflowManager.execute_from` is called with no argument to get the origin nodes.
        """

        # This create_all is supposed to only create workflow-management side tables (called "wom_*")
        SQLManager.instance().create_all()

        if OptionManager.instance()["--clear-history"]:
            Logger.instance().info("Deleting WoPMaRS history...")
            SQLManager.instance().drop_table_content_list(SQLManager.wom_table_names)

        # The following lines allow to create types 'input' and 'output' in the db if they don't exist.
        self.__session.get_or_create(Type, defaults={"id": 1}, name="input")
        self.__session.get_or_create(Type, defaults={"id": 2}, name="output")
        self.__session.commit()
        # Get the DAG representing the whole workflow
        self.__dag_tools = self.__parser.parse()
        # Build the DAG which is willing to be executed according
        self.get_dag_to_exec()
        # Start the execution at the root nodes
        if OptionManager.instance()["--forceall"] and not OptionManager.instance()["--dry-run"]:
            self.erase_output()
        self.execute_from()

    def erase_output(self):
        """
        Erase the outputs of the DAG that will be executed in order to prevents conflicts.
        """
        list_tw = self.__dag_to_exec.nodes()
        set_files = set()
        set_tables = set()

        Logger.instance().info("Forced execution implies overwrite existing output. Erasing files and tables.")
        for tw in list_tw:
           [set_files.add(f.path) for f in tw.files if f.type.name == "output"]
           [set_tables.add(t.tablename) for t in tw.tables if t.type.name == "output"]

        s = ""
        for f_path in set_files:
            s += "\n" + f_path
            PathFinder.silentremove(f_path)
        Logger.instance().debug("Removed files:" + s)

        SQLManager.instance().drop_table_content_list(
            set(IODbPut.tablenames).intersection(set_tables))

        s = "\n"
        s += "\n".join(set_tables)
        Logger.instance().debug("Removed tables content:" + s)

        Logger.instance().info("Output files and tables from previous execution have been erased.")

    def get_dag_to_exec(self):
        """
        Set the dag to exec in terms of --sourcerule option and --targetrule option.

        The source rule is checked first (there should not be both set because of the checks at the begining of the software)

        If sourcerule is set, then it is its successors that are searched in the whole dag.
        Else, it is its predecessors.

        The set of obtained rules are used to build the "dag_to_exec". The nodes returned by get_all_successors and
        get_all_predecessors are implicitly all related.
        """
        if OptionManager.instance()["--sourcerule"] is not None:
            try:
                # Get the rule asked by the user as 'sourcerule'
                node_from_rule = [n for n in self.__dag_tools if n.name == OptionManager.instance()["--sourcerule"]][0]
            except IndexError:
                raise WopMarsException(
                    "The given rule to start from: " + OptionManager.instance()["--sourcerule"] + " doesn't exist.")

            self.__dag_to_exec = DAG(self.__dag_tools.get_all_successors(node_from_rule))
            Logger.instance().info("Running the workflow from rule " + str(OptionManager.instance()["--sourcerule"]) +
                                   " -> " + node_from_rule.toolwrapper)
        elif OptionManager.instance()["--targetrule"] is not None:
            try:
                # Get the rule asked by the user as 'targetrule'
                node_from_rule = [n for n in self.__dag_tools if n.name == OptionManager.instance()["--targetrule"]][0]
            except IndexError:
                raise WopMarsException(
                    "The given rule to go to: " + OptionManager.instance()["--targetrule"] + " doesn't exist.")
            self.__dag_to_exec = DAG(self.__dag_tools.get_all_predecessors(node_from_rule))
            Logger.instance().info("Running the workflow to the rule " + str(OptionManager.instance()["--targetrule"]) +
                                   " -> " + node_from_rule.toolwrapper)
        else:
            self.__dag_to_exec = self.__dag_tools

        # ???
        # todo checkout what is going on here
        tables = []
        [tables.extend(tw.tables) for tw in self.__dag_to_exec.nodes()]
        IODbPut.set_tables_properties(tables)

        # For the tools that are in the workflow definition file but not in the executed dag, their status is set to
        # "NOT_PLANNED"
        for tw in set(self.__dag_tools.nodes()).difference(set(self.__dag_to_exec.nodes())):
            tw.set_execution_infos(status="NOT_PLANNED")
            self.__session.add(tw)

        self.__session.commit()

    def execute_from(self, tw=None):
        """
        Execute the dag from the toolwrappers in the list given.

        The next nodes are taken thanks to the "successors()" method of the DAG and are put into the queue.
        The "run_queue()" is then called.

        A trace of the already_runned ToolWrapper objects is kept in order to avoid duplicate execution.

        :param node: A node of the DAG or None, if it needs to be executed from the root.
        :type node: :class:`~.wopmars.framework.database.tables.ToolWrapper.ToolWrapper`
        :return: void
        """
        # the first list will be the root nodes
        list_tw = self.__dag_to_exec.successors(tw)
        Logger.instance().debug("Next tools: " + str([t.__class__.__name__ for t in list_tw]))

        for tw in list_tw:
            # every rule should be executed once and only once
            if tw not in self.__already_runned:
                # ToolThread object is a thread ready to start
                self.__queue_exec.put(ToolThread(tw))
            else:
                Logger.instance().debug("Rule: " + tw.name +
                                        " -> " + tw.toolwrapper +
                                        " has already been executed. Pass.")
        self.run_queue()

    def run_queue(self):
        """
        Call start() method of all elements of the queue.

        The tools inside the queue are taken then their inputs are checked. If they are ready, the tools are started.
        If not, they are put in a buffer list of "not ready tools" or "ready but has not necessary ressources available
        tools".

        The start method is called with a dry argument, if it appears that the input of the ToolWrapper are the same
        than in a previous execution, and that the output are already ready. The dry parameter is set to True and the
        start method will only simulate the execution.

        After that, the code check for the state of the workflow and gather the informations to see if the workflow
        is finished, if it encounter an error or if it is currently running.

        :raises WopMarsException: The workflow encounter a problem and must stop.
        """

        #
        # # TODO THIS METHOD IS NOT THREAD-SAFE (peut etre que si, à voir)
        #

        # If no tools have been added to the queue:
        #  - All tools have been executed and the queue is empty, so nothing happens
        #  - There were remaing tools in the queue but they weren't ready, so they are tested again
        while not self.__queue_exec.empty():
            Logger.instance().debug("Queue size: " + str(self.__queue_exec.qsize()))
            Logger.instance().debug("Queue content: " + str(["rule: " + tt.get_toolwrapper().name + "->" +
                                                             tt.get_toolwrapper().toolwrapper for tt in self.__queue_exec.get_queue_tuple()]))
            # get the first element of the queue to execute
            thread_tw = self.__queue_exec.get()
            tw = thread_tw.get_toolwrapper()
            Logger.instance().debug("Current rule: " + tw.name + "->" + tw.toolwrapper)
            # check if the predecessors of a rule have been already executed: a rule shouldn't be executed if
            # its predecessors have not been executed yet
            if not self.all_predecessors_have_run(tw):
                Logger.instance().debug("Predecessors of rule: " + tw.name + " have not been executed yet.")
            # for running, either the inputs have to be ready or the dry-run mode is enabled
            elif tw.are_inputs_ready() or OptionManager.instance()["--dry-run"]:
                # the state of inputs (table and file) are set in the db here.
                tw.set_args_date_and_size("input")
                Logger.instance().debug("ToolWrapper ready: " + tw.toolwrapper)
                dry = False
                # if forceall option, then the tool is reexecuted anyway
                # check if the actual execution of the toolwrapper is necessary
                # every predecessors of the toolwrapper have to be executed (or simulated)
                if not OptionManager.instance()["--forceall"] and \
                        self.is_this_tool_already_done(tw) and \
                        not bool([node for node in self.__dag_to_exec.predecessors(tw) if node.status != "EXECUTED" and
                                        node.status != "ALREADY_EXECUTED"]):
                    Logger.instance().info("Rule: " + tw.name + " -> " + tw.toolwrapper +
                                           " seemed to have already" +
                                           " been runned with same" +
                                           " parameters.")
                    dry = True

                # todo twthread verification des ressources
                thread_tw.subscribe(self)
                self.__count_exec += 1
                # todo twthread methode start
                thread_tw.set_dry(dry)
                try:
                    # be carefull here: the execution of the toolthreads is recursive meaning that calls to function may
                    # be stacked (run -> notify success -> run(next tool) -> notify success(next tool) -> etc....
                    # todo twthread methode start
                    thread_tw.run()
                except Exception as e:
                    # as mentionned above, there may be recursive calls to this function, so every exception can
                    # pass here multiple times: this attribute is used for recognizing exception that have already been
                    # caught
                    if not hasattr(e, "teb_already_seen"):
                        setattr(e, "teb_already_seen", True)
                        tw.set_execution_infos(status="EXECUTION_ERROR")
                        self.__session.add(tw)
                        self.__session.commit()
                    raise e
            else:
                Logger.instance().debug("ToolWrapper not ready: rule: " + tw.name + " -> " + str(tw.toolwrapper))
                # The buffer contains the ToolWrappers that have inputs which are not ready yet.
                self.__list_queue_buffer.append(thread_tw)

        Logger.instance().debug("Buffer: " + str(["rule: " + t.get_toolwrapper().name + "->" +
                                                  t.get_toolwrapper().toolwrapper for t in self.__list_queue_buffer]))
        Logger.instance().debug("Running rules: " + str(self.__count_exec))

        # There is no more ToolWrapper that are waiting to be executed.
        # Is there some tools that are currently being executed?
        if self.__count_exec == 0:
            # Is there some tools that weren't ready?
            if len(self.__list_queue_buffer) == 0:
                # If there is no tool waiting and no tool being executed, the workflow has finished.
                finished_at = datetime.datetime.fromtimestamp(time.time())
                Logger.instance().info("The workflow has completed. Finished at: " + str(finished_at))
                self.set_finishing_informations(finished_at, "FINISHED")
                SQLManager.instance().get_session().close()
                sys.exit(0)
            # uniquement en environnement multiThreadpredece
            elif not self.check_buffer():
                # If there is no tool being executed but there is that are waiting something, the workflow has an issue
                finished_at = datetime.datetime.fromtimestamp(time.time())
                tw_list = [t.get_toolwrapper() for t in self.__list_queue_buffer]
                if len(tw_list) > 0:
                    input_files_not_ready = tw_list[0].get_input_files_not_ready()
                    self.set_finishing_informations(finished_at, "ERROR")
                    raise WopMarsException("The workflow has failed.",
                                           "The inputs '{}' have failed for this tool '{}'".format(input_files_not_ready[0], tw_list[0].name))
                                           # "The inputs are not ready for thisto: " +
                                           # ", \n".join([t.get_toolwrapper().toolwrapper +
                                           #            " -> rule: " +
                                           #            t.get_toolwrapper().name for t in self.__list_queue_buffer]) + ". ")
            # If there is one tool that is ready, it means that it is in queue because ressources weren't available.

    def set_finishing_informations(self, finished_at, status):
        """
        Set the finsihing information of the whole workflow.

        :param finished_at: The finishing date of the workflow
        :type finished_at: datetime.datetime
        :param status: The final status of the workflow
        :type status: str
        """
        modify_exec = self.__session.query(Execution).order_by(Execution.id.desc()).first()
        if modify_exec is not None:
            modify_exec.finished_at = finished_at
            modify_exec.time = (modify_exec.finished_at - modify_exec.started_at).total_seconds()
            modify_exec.status = status
            self.__session.add(modify_exec)
            self.__session.commit()

    def all_predecessors_have_run(self, tw):
        """
        Check if all the predecessors of the given toolwrapper have yet been executed in this workflow.

        :param tw: Node of the DAG
        :type tw: :class:`~.wopmars.main.framework.database.tables.ToolWrapper.ToolWrapper`
        :return: Bool
        """
        return bool(self.__dag_to_exec.get_all_predecessors(tw).difference(set([tw])).issubset(set(self.__already_runned)))

    @staticmethod
    def is_this_tool_already_done(tw):
        """
        Return True if conditions for saying "The output of this ToolWrapper are already available" are filled.

        The conditions are:
            - The ToolWrapper exist in database (named = tw_old)
            - The tw_old param are the same than the same which is about to start
            - the tw_old inputs are the same
            - the tw_old outputs exists with the same name and are more recent than inputs

        :param tw: The Toolwrapper to test_bak
        :type tw: :class:`~.wopmars.main.framework.database.tables.ToolWrapper.ToolWrapper`
        """
        session = SQLManager.instance().get_session()
        list_same_toolwrappers = session.query(ToolWrapper).filter(ToolWrapper.toolwrapper == tw.toolwrapper)\
            .filter(ToolWrapper.execution_id != tw.execution_id).all()
        i = 0
        while i < len(list_same_toolwrappers):
            same = False
            # two tw are equals if they have the same parameters, the same file names and path
            # and the same table names and models
            if list_same_toolwrappers[i] == tw and \
                    list_same_toolwrappers[i].does_output_exist() and \
                    list_same_toolwrappers[i].is_output_more_recent_than_input():
                    same = True
            if not same:
                del list_same_toolwrappers[i]
            else:
                i += 1

        # The elements of the list have been removed if none fit the conditions
        return bool(list_same_toolwrappers)

    def check_buffer(self):
        """
        Check if the buffer contains ToolWrapper that are ready.

        :return: bool: True if there is at least one toolwrapper that is READY in the buffer.
        """
        for tw in self.__list_queue_buffer:
            if tw.get_toolwrapper().get_state() == ToolWrapper.READY:
                return True
        return False

    def notify_success(self, thread_toolwrapper):
        """
        Handle thread_toolwrapper success by continuing the dag.

        :param thread_toolwrapper: ToolWrapper thread that just succeed
        :type thread_toolwrapper: :class:`~.wopmars.management.ToolThread.ToolThread`
        """
        self.__session.add(thread_toolwrapper.get_toolwrapper())
        self.__session.commit()

        dry_status = thread_toolwrapper.get_dry()
        # if not OptionManager.instance()["--dry-run"]:
        #     thread_toolwrapper.get_toolwrapper().set_args_date_and_size("output", dry_status)
        if dry_status is False and not OptionManager.instance()["--dry-run"]:
            Logger.instance().info("Rule " + str(thread_toolwrapper.get_toolwrapper().name) + " -> " + str(thread_toolwrapper.get_toolwrapper().__class__.__name__) + " has succeed.")
        # Continue the dag execution from the toolwrapper that just finished.
        self.__already_runned.add(thread_toolwrapper.get_toolwrapper())
        self.__count_exec -= 1

        if len(self.__list_queue_buffer):
            Logger.instance().debug("Fill the queue with the Buffer: " +
                                    str([t.get_toolwrapper().__class__.__name__ for t in self.__list_queue_buffer]))
        i = 0
        for tw_thread in self.__list_queue_buffer:
            self.__queue_exec.put(tw_thread)
            del self.__list_queue_buffer[i]
            i += 1

        self.execute_from(thread_toolwrapper.get_toolwrapper())

    def notify_failure(self, thread_toolwrapper):
        """
        Handle thread_toolwrapper failure by re-puting it in the queue.

        :param thread_toolwrapper: ToolWrapper thread that just failed
        :return:
        """
        pass



from wopmars.framework.database.tables.ToolWrapper import ToolWrapper


class SparePartsManufacturer(ToolWrapper):
    __mapper_args__ = {
        "polymorphic_identity": "wopmars.example.SparePartsManufacturer"
    }

    def specify_input_file(self):
        return ["pieces"]

    def specify_output_table(self):
        return ["piece"]

    def specify_params(self):
        return {
            "max_price": int
        }

    def run(self):
        session = self.session()
        Piece = self.output_table("piece")

        with open(self.input_file("pieces")) as wr:
            lines = wr.readlines()

        for line in lines:
            splitted_line = line.split(";")
            piece_serial_number = splitted_line[0]
            piece_type = splitted_line[1]
            piece_price = float(splitted_line[2])
            if (self.option("max_price") and self.option("max_price" >= piece_price)) \
                    or self.option("max_price") is None:
                session.add(Piece(serial_number=piece_serial_number, price=piece_price, type=piece_type))

        session.commit()



"""
Module containing the FooWrapper1 class
"""
import os
import time

from matplotlib.compat import subprocess

from wopmars.framework.database.tables.ToolWrapper import ToolWrapper


class FooWrapperRelationShip(ToolWrapper):
    """
    This class has been done for example/testing purpose.
    Modifications may lead to failure in tests.
    """
    __mapper_args__ = {'polymorphic_identity': "FooWrapperRelationShip"}

    def specify_output_table(self):
        return ["FooBase3", "FooBase4", "FooBase5", "FooBase6"]

    def run(self):
        for i in range(10):
            f3 = self.output_table("FooBase3")(id=i+1, name="F3 - " + str(i), id_foobase6=i+1)
            f4 = self.output_table("FooBase4")(id_foobase3=i+1, id_foobase5=i+1)
            f5 = self.output_table("FooBase5")(id=i+1)
            f6 = self.output_table("FooBase6")(id=i+1)
            self.session().add_all([f3, f4, f5, f6])
        self.session().commit()


from sqlalchemy import Column, Integer

from wopmars.framework.database.Base import Base


class Failure(Base):
    __tablename__ = "BaseWrongClassName"

    id = Column(Integer, primary_key=True, autoincrement=True)




"""
This module contains the WopMarsException class
"""


class WopMarsException(Exception):
    """
    WopMarsException inherit from Exception and is thrown by every part of the .

    This exception is thrown with a context and a message. The context is supposed to explain where the exception has
    occured while the message gives more specific details about the exception.
    """

    def __init__(self, context="", message=""):
        """
        :param context: String that contains the context
        :param message: string that contains the precise error
        :return: void
        """
        self.__context = context
        self.__message = message

    def __str__(self):
        s = ""
        if self.__context:
            s += self.__context + "\n\t\t"
        if self.__message:
            s += "\n\t\t".join(self.__message.split("\n"))
        return s



import unittest
from unittest import TestCase

from wopmars.framework.database.tables.IOFilePut import IOFilePut
from wopmars.utils.SetUtils import SetUtils


class TestSetUtils(TestCase):
    def setUp(self):
        # 1 in 2 but not 2 in 1
        self.__set1 = set("a")
        self.__set2 = set(("a", "b"))

    def test_all_elm_of_one_set_in_one_other(self):
        self.assertTrue(SetUtils.all_elm_of_one_set_in_one_other(self.__set1, self.__set2))
        self.assertFalse(SetUtils.all_elm_of_one_set_in_one_other(self.__set2, self.__set1))

if __name__ == "__main__":
    unittest.main()


"""
Module containing the Singleton function.
"""
import threading


def singleton(cls):
    instances = {}

    def class_instanciation_or_not(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]
    return class_instanciation_or_not


# Based on tornado.ioloop.IOLoop.instance() approach.
# See https://github.com/facebook/tornado
class SingletonMixin(object):
    __singleton_lock = threading.Lock()
    __singleton_instance = None

    @classmethod
    def instance(cls):
        if not cls.__singleton_instance:
            with cls.__singleton_lock:
                if not cls.__singleton_instance:
                    cls.__singleton_instance = cls()
        return cls.__singleton_instance

    @classmethod
    def _drop(cls):
        """Drop the instance (for testing purposes)."""
        if cls.__singleton_instance:
            with cls.__singleton_lock:
                if cls.__singleton_instance:
                    cls.__singleton_instance = None


"""
This module contains the DictUtils class.
"""


class DictUtils:
    """
    Class containing methods that are usefull for dealing with dicts
    """
    @staticmethod
    def elm_of_one_dict_in_one_other(one, other):
        """
        Check if all elements of one dict is in an other.

        :param one: dict
        :param other: dict
        :return: boolean: True if all elements of one are in the other
        """
        if len(set(one.keys()).difference(set(other.keys()))):
            return False
        for elm in one:
            if one[elm] != other[elm]:
                return False
        return True

    @staticmethod
    def at_least_one_value_of_one_in_an_other(one, other):
        for value1 in one.values():
            for value2 in other.values():
                if value1 == value2:
                    return True
        return False

    @staticmethod
    def pretty_repr(d, indent=0):
        s = ""
        for key, value in d.items():
            s += '\t' * indent + str(key) + "\n"
            if isinstance(value, dict):
                s += DictUtils.pretty_repr(value, indent + 1)
            else:
                s += '\t' * (indent + 1) + str(value) + "\n"
        return s

    @staticmethod
    def str_to_dict(dictable_string):
        if dictable_string:
            d = dict(eval(dictable_string))
            return d



import unittest
from unittest import TestCase

from test.resource.wrapper.FooWrapper2 import FooWrapper2
from test.resource.wrapper.FooWrapper8 import FooWrapper8
from wopmars.framework.database.SQLManager import SQLManager
from wopmars.framework.database.tables.IOFilePut import IOFilePut
from wopmars.framework.database.tables.ToolWrapper import ToolWrapper
from wopmars.framework.database.tables.Type import Type
from wopmars.framework.management.DAG import DAG
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.PathFinder import PathFinder


class TestDAG(TestCase):

    def setUp(self):
        OptionManager.initial_test_setup()
        SQLManager.instance().create_all()
        #        first
        #       /    \
        #   second   third
        #       \    /
        #       fourth
        #

        self.__session = SQLManager.instance().get_session()

        input_entry = Type(name="input")
        output_entry = Type(name="output")

        f1 = IOFilePut(name="input1", path="file1.txt")
        f1.type = input_entry

        f2 = IOFilePut(name="output1", path="file2.txt")
        f2.type = output_entry

        self.__toolwrapper_first = FooWrapper2(rule_name="rule1")
        self.__toolwrapper_first.files.extend([f1, f2])

        f1 = IOFilePut(name="input1", path="file2.txt")
        f1.type = input_entry

        f2 = IOFilePut(name="output1", path="file3.txt")
        f2.type = output_entry

        self.__toolwrapper_second = FooWrapper2(rule_name="rule2")
        self.__toolwrapper_second.files.extend([f1, f2])

        f1 = IOFilePut(name="input1", path="file2.txt")
        f1.type = input_entry

        f2 = IOFilePut(name="output1", path="file4.txt")
        f2.type = output_entry

        self.__toolwrapper_third = FooWrapper2(rule_name="rule3")
        self.__toolwrapper_third.files.extend([f1, f2])

        f1 = IOFilePut(name="input1", path="file3.txt")
        f1.type = input_entry

        f2 = IOFilePut(name="input2", path="file4.txt")
        f2.type = input_entry

        f3 = IOFilePut(name="output1", path="file5.txt")
        f3.type = output_entry

        self.__toolwrapper_fourth = FooWrapper8(rule_name="rule4")
        self.__toolwrapper_fourth.files.extend([f1, f2, f3])

        list_tool = [self.__toolwrapper_first,
                     self.__toolwrapper_second,
                     self.__toolwrapper_third,
                     self.__toolwrapper_fourth]

        self.__set_tool = set(list_tool)

        SQLManager.instance().get_session().add_all(list_tool)
        SQLManager.instance().get_session().commit()

    def test_get_all_successors(self):
        my_dag = DAG(self.__set_tool)

        self.assertEqual(set(my_dag.get_all_successors(self.__toolwrapper_first)), self.__set_tool)
        self.assertNotEqual(set(my_dag.get_all_successors(self.__toolwrapper_first)),
                            self.__set_tool.difference(set([self.__toolwrapper_second])))
        self.assertNotEqual(set(my_dag.get_all_successors(self.__toolwrapper_first)),
                            self.__set_tool.difference(set([self.__toolwrapper_first])))

    def test_get_all_predecessors(self):
        my_dag = DAG(self.__set_tool)

        self.assertEqual(set(my_dag.get_all_predecessors(self.__toolwrapper_fourth)), self.__set_tool)
        self.assertNotEqual(set(my_dag.get_all_predecessors(self.__toolwrapper_fourth)),
                            self.__set_tool.difference(set([self.__toolwrapper_second])))
        self.assertNotEqual(set(my_dag.get_all_predecessors(self.__toolwrapper_fourth)),
                            self.__set_tool.difference(set([self.__toolwrapper_fourth])))

    def tearDown(self):
        SQLManager.instance().get_session().close() 
        SQLManager.instance().drop_all()
        OptionManager._drop()
        PathFinder.silentremove("test/output/output_file1.txt")
        SQLManager._drop()

    def test_init(self):
        try:
            my_dag = DAG(self.__set_tool)

            dag_from_base = DAG(set(SQLManager.instance().get_session().query(ToolWrapper).all()))
            self.assertEqual(my_dag, dag_from_base)

            # Verifying that the nodes are correctly sorted
            self.assertTrue(self.__toolwrapper_fourth in my_dag.successors(self.__toolwrapper_third))
            self.assertTrue(self.__toolwrapper_fourth in my_dag.successors(self.__toolwrapper_second))
            self.assertTrue(self.__toolwrapper_second in my_dag.successors(self.__toolwrapper_first))
            self.assertTrue(self.__toolwrapper_third in my_dag.successors(self.__toolwrapper_first))
        except:
            # beware, if a test_bak inside the try block fails, 2 exceptions are raised
            raise AssertionError('Should not raise exception')

if __name__ == '__main__':
    unittest.main()




import datetime
import importlib
#import jinja2
import re
import time
import os

import yaml
from yaml.constructor import ConstructorError
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from sqlalchemy.orm.exc import NoResultFound, ObjectDeletedError

from wopmars.framework.database.SQLManager import SQLManager
from wopmars.framework.database.tables.Execution import Execution
from wopmars.framework.database.tables.IODbPut import IODbPut
from wopmars.framework.database.tables.IOFilePut import IOFilePut
from wopmars.framework.database.tables.ModificationTable import ModificationTable
from wopmars.framework.database.tables.Option import Option
from wopmars.framework.database.tables.Type import Type
from wopmars.utils.DictUtils import DictUtils

from wopmars.utils.Logger import Logger
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.exceptions.WopMarsException import WopMarsException


class Reader:
    """
    This class is responsible of the parsing of the user's entries:

     - the arguments given with the ``tool`` command
     - the content of the workflow definition file, in normal mode

    This module uses the ``yaml`` library in order to parse the workflow definition file. Some additional rules have been
    added to the ``yaml`` library in order to prevent duplicate rules.

    Also, once the ``Reader`` has gotten the workflow definition informations, it'll check for eventual errors and then
    store them in the database. Those stored informations are what we call the "history" of **WoPMaRS**.
    """
    def __init__(self):
        self.__dict_workflow_definition = None

    def load_definition_file(self, s_definition_file):
        """
        Open the definition file and load it's content in a dictionnary thanks to the ``yaml`` library. ``yaml`` can
        raise an exception if the yaml specifications are not respected or if there is duplicates at the same level of
        hierarchy in the definition file. If so, the exception is caught then wrapped into a ``WopMarsException``.

        The check of the grammar of the definition file is done during this step but no tests are performed regarding
        to the actual content of the definition file.

        :param s_definition_file: Path to the definition file
        :type s_definition_file: str
        :raises WopMarsException: The yaml specifications are not respected
        """
        # Tests about grammar and syntax are performed here (file's existence is also tested here)
        try:
            with open(s_definition_file, 'r') as def_file:
                s_def_file_content = def_file.read()
            try:
                # The workflow definition file is loaded as-it in memory by the pyyaml library
                Logger.instance().info("Reading the Wopfile: " + str(s_definition_file))
                # Replace jinja2 variables with environment variable values
                #s_def_file_content = jinja2.Environment().from_string(s_def_file_content).render(os.environ)
                # Parse the file to find duplicates rule names (it is a double check with the following step)
                Reader.check_duplicate_rules(s_def_file_content)
                # Allows to raise an exception if duplicate keys are found on the same document hirearchy level.
                yaml.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, Reader.no_duplicates_constructor)
                # The whole content of the definition file is loaded in this dict.
                # yaml.load return None if there is no content in the String
                self.__dict_workflow_definition = yaml.load(s_def_file_content) or {}
                if self.__dict_workflow_definition == {}:
                    Logger.instance().warning("The workflow definition file is empty")
                Logger.instance().debug("\n" + DictUtils.pretty_repr(self.__dict_workflow_definition))
                Logger.instance().debug("Read complete.")
                Logger.instance().info("Checking whether the file is well formed...")
                # raise an exception if there is a problem with the grammar
                self.is_grammar_respected()
                Logger.instance().debug("File well formed.")
            # YAMLError is thrown if the YAML specifications are not respected by the definition file
            except yaml.YAMLError as exc:
                raise WopMarsException("Error while parsing the configuration file: \n\t"
                                       "The YAML specification is not respected:", str(exc))
            except ConstructorError as CE:
                raise WopMarsException("Error while parsing the configuration file: \n\t",
                                       str(CE))
        except FileNotFoundError:
            raise WopMarsException("Error while parsing the configuration file: \n\tInput error:",
                                   "The specified file at " + s_definition_file + " doesn't exist.")

    # Code from the github: https://gist.github.com/pypt/94d747fe5180851196eb
    @staticmethod
    def no_duplicates_constructor(loader, node, deep=False):
        """
        Make the yaml constructor to check for duplicate keys.
        """
        mapping = {}
        for key_node, value_node in node.value:
            key = loader.construct_object(key_node, deep=deep)
            value = loader.construct_object(value_node, deep=deep)
            if key in mapping:
                raise ConstructorError("while constructing a mapping", node.start_mark,
                                       "found duplicate key (%s)" % key, key_node.start_mark)
            mapping[key] = value
        return loader.construct_mapping(node, deep)

    @staticmethod
    def check_duplicate_rules(s_workflow_file):
        """
        This method raises an exception if the workflow definition file contains duplicate rule names.

        The workflow definition file should contain rules with different name. It is therefore recommended to not
        call rules with tool names but functionality instead. Example:

            .. code-block:: yaml

                rule get_snp:
                    tool: SNPGetter
                    input:
                        file:
                            etc..
                        table:
                            etc..
                    output:
                        file:
                            etc..
                        table:
                            etc..
                    params:
                        etc..

        :param s_workflow_file: The content of the definition file
        :type s_workflow_file: str
        :raises WopMarsException: There is a duplicate rule name
        """
        Logger.instance().debug("Looking for duplicate rules...")
        # All rules are found using this regex.
        rules = re.findall(r'rule (.+?):', str(s_workflow_file))
        seen = set()
        # for each rule name
        for r in rules:
            # if the rule has not been seen before
            if r not in seen:
                # add it to the set of seen rules
                seen.add(r)
            else:
                # There is a duplicate rule name
                raise WopMarsException("Error while parsing the configuration file:\n\t",
                                       "The rule " + r + " is duplicated.")
        Logger.instance().debug("No Duplicate.")

    def is_grammar_respected(self):
        """
        Check if the definition file respects the grammar. Throw a WopMarsException exception if not.

        The formal representation of the grammar is::

            WoPMaRS       = rule
            identifier    = String
            ni            = NEWLINE INDENT
            rule          = "rule" identifier ":" ruleparams
            ruleparams    = [ni tool] [ni input] [ni output] [ni params]
            filesortables = (ni files|ni tables){0-2}
            files         = "file"  ":" (ni identifier ”:” stringliteral)+
            tables        = "table"  ":" (ni identifier ”:” stringliteral)+
            tool          = "tool"   ":" stringliteral
            input         = "input"  ":" ni filesortables
            output        = "output" ":" ni filesortables
            params        = "params" ":" (ni identifier ”:” stringliteral)+
            (NEWLINE WoPMaRS)+

        :raises WopMarsException: The grammar is not respected
        """
        exemple_file_def = """
    rule RULENAME:
        tool: TOOLNAME
        input:
            file:
                INPUTNAME: INPUTVALUE
            table:
                - path.to.table
        output:
            file:
                OUTPUTNAME: OUTPUTVALUE
            table:
                - path.to.table
        params:
            OPTIONNAME: OPTIONVALUE

    rule ...etc...
        """
        # recognize the rule blocks
        regex_step1 = re.compile(r"(^rule [^\s]+$)")

        # recognize the elements of the rule
        regex_step2 = re.compile(r"(^params$)|(^tool$)|(^input$)|(^output$)")

        # recognize the file/table blocks
        regex_step3 = re.compile(r"(^file$)|(^table$)")

        # The words found are tested against the regex to see if they match or not
        for s_key_step1 in self.__dict_workflow_definition:
            bool_toolwrapper = False
            # The first level of indentation should only contain rules
            if not regex_step1.search(s_key_step1):
                raise WopMarsException("Error while parsing the configuration file: \n\t"
                                       "The grammar of the WopMars's definition file is not respected:",
                                       "The line containing:\'" +
                                       str(s_key_step1) +
                                       "\' doesn't match the grammar: it should start with 'rule'" +
                                       "and contains only one word after the 'rule' keyword" +
                                       "\nexemple:" + exemple_file_def)

            for s_key_step2 in self.__dict_workflow_definition[s_key_step1]:
                # the second level of indentation should only contain elements of rule
                if not regex_step2.search(s_key_step2):
                    raise WopMarsException("Error while parsing the configuration file: \n\t"
                                           "The grammar of the WopMars's definition file is not respected:",
                                           "The line containing:'" + str(s_key_step2) + "'" +
                                           " for rule '" + str(s_key_step1) + "'" +
                                           " doesn't match the grammar: it should be " +
                                           "'tool', 'params', 'input' or 'output'" +
                                           "\nexemple:" + exemple_file_def)
                elif s_key_step2 == "input" or s_key_step2 == "output":
                    for s_key_step3 in self.__dict_workflow_definition[s_key_step1][s_key_step2]:
                        if not regex_step3.search(s_key_step3):
                            raise WopMarsException("Error while parsing the configuration file: \n\t"
                                                   "The grammar of the WopMars's definition file is not respected:",
                                                   "The line containing:'" + str(s_key_step3) + "'" +
                                                   " for rule '" + str(s_key_step1) + "'" +
                                                   " doesn't match the grammar: it should be " +
                                                   "'file' or 'table'" +
                                                   "\nexemple:" + exemple_file_def)
                        elif s_key_step3 == "file":
                            for s_variable_name in self.__dict_workflow_definition[s_key_step1][s_key_step2][s_key_step3]:
                                if type(self.__dict_workflow_definition[s_key_step1][s_key_step2][s_key_step3][s_variable_name]) != str:
                                    raise WopMarsException("Error while parsing the configuration file: \n\t" +
                                                           "The grammar of the WopMars's definition file is not respected:",
                                                           "The line containing:'" + str(s_variable_name) + "'" +
                                                           " for rule '" + str(s_key_step1) + "'" +
                                                           " doesn't match the grammar: it should be the string containing the path to the file."
                                                           "\nexemple:" + exemple_file_def)
                        elif s_key_step3 == "table":
                            for s_tablename in self.__dict_workflow_definition[s_key_step1][s_key_step2][s_key_step3]:
                                if type(s_tablename) != str:
                                    raise WopMarsException("Error while parsing the configuration file: \n\t"
                                                           "The grammar of the WopMars's definition file is not respected:",
                                                           "The line containing:'" + str(s_variable_name) + "'" +
                                                           " for rule '" + str(s_key_step1) + "'" +
                                                           " doesn't match the grammar: it should be the string containing the name of the Model."
                                                           "\nexemple:" + exemple_file_def)

                # There should be one tool at max in each rule
                elif s_key_step2 == "tool":
                    if bool_toolwrapper == False:
                        bool_toolwrapper = True
                    elif bool_toolwrapper == True:
                        raise WopMarsException("Error while parsing the configuration file: \n\t",
                                               "There is multiple tools specified for the " + str(s_key_step1))

            # All rules should contain a tool
            if not bool_toolwrapper:
                raise WopMarsException("Error while parsing the configuration file: \n\t"
                                       "The grammar of the WopMars's definition file is not respected:",
                                       "The rule '" + str(s_key_step1) + "' doesn't contain any tool." +
                                       "\nexemple:" + exemple_file_def
                                       )

    def load_one_toolwrapper(self, s_toolwrapper, s_dict_inputs, s_dict_outputs, s_dict_params):
        """
        Method called when the ``tool`` command is used. It is equivalent to the :meth:`~.wopmars.framework.parsing.Reader.Reader.read` method but create a workflow
        with only one toolwrapper. The workflow is also stored inside the database.

        :param s_toolwrapper: The name of the toolwrapper (will be imported)
        :type s_toolwrapper: str
        :param s_dict_inputs: A string containing the dict of input files
        :type s_dict_inputs: str
        :param s_dict_outputs: A string containing the dict of output files
        :type s_dict_outputs: str
        :param s_dict_params: A string containing the dict of params
        :type s_dict_params: str

        :raise WopMarsException: There is an error while accessing the database
        """
        session = SQLManager.instance().get_session()
        dict_inputs = dict(eval(s_dict_inputs))
        dict_outputs = dict(eval(s_dict_outputs))
        dict_params = dict(eval(s_dict_params))
        try:
            # The same execution entry for the whole workflow-related database entries.
            execution = Execution(started_at=datetime.datetime.fromtimestamp(time.time()))
            # get the types that should have been created previously
            input_entry = session.query(Type).filter(Type.name == "input").one()
            output_entry = session.query(Type).filter(Type.name == "output").one()

            Logger.instance().debug("Loading unique toolwrapper " + s_toolwrapper)
            dict_dict_dict_elm = dict(dict_input={"file": {}, "table": {}},
                                 dict_params={},
                                 dict_output={"file": {}, "table": {}})
            for type in dict_inputs:
                if type == "file":
                    for s_input in dict_inputs[type]:
                        obj_created = IOFilePut(name=s_input,
                                                path=os.path.abspath(os.path.join(OptionManager.instance()["--directory"],
                                                                                  dict_inputs[type][s_input])))
                        dict_dict_dict_elm["dict_input"][type][s_input] = obj_created
                        Logger.instance().debug("Object input file: " + s_input + " created.")
                elif type == "table":
                    for s_input in dict_inputs[type]:
                        obj_created = IODbPut(model=dict_inputs[type][s_input],
                                              tablename=s_input)
                        dict_dict_dict_elm["dict_input"][type][s_input] = obj_created
                        Logger.instance().debug("Object input table: " + s_input + " created.")
            for type in dict_outputs:
                if type == "file":
                    for s_output in dict_outputs[type]:
                        obj_created = IOFilePut(name=s_output,
                                                path=os.path.abspath(os.path.join(OptionManager.instance()["--directory"],
                                                                                  dict_outputs[type][s_output])))
                        dict_dict_dict_elm["dict_output"]["file"][s_output] = obj_created
                        Logger.instance().debug("Object output file: " + s_output + " created.")
                elif type == "table":
                    for s_output in dict_outputs[type]:
                        obj_created = IODbPut(model=dict_outputs[type][s_output],
                                              tablename=s_output)
                        dict_dict_dict_elm["dict_output"]["table"][s_output] = obj_created
                        Logger.instance().debug("Object output table: " + s_output + " created.")
            for s_param in dict_params:
                obj_created = Option(name=s_param,
                                     value=dict_params[s_param])
                dict_dict_dict_elm["dict_params"][s_param] = obj_created
                Logger.instance().debug("Object option: " + s_param + " created.")

            # Instantiate the refered class
            wrapper_entry = self.create_toolwrapper_entry("rule_" + s_toolwrapper, s_toolwrapper,
                                                          dict_dict_dict_elm, input_entry, output_entry)
            wrapper_entry.execution = execution
            Logger.instance().debug("Object toolwrapper: " + s_toolwrapper + " created.")
            session.add(wrapper_entry)
            session.commit()
            session.rollback()
            IODbPut.set_tables_properties(IODbPut.get_execution_tables())
            # commit /rollback trick to clean the session
            # todo ask lionel est-ce-que tu as deja eu ce problème à ne pas pouvoir faire des queries et des ajouts
            # dans la meme session?
            session.commit()
            session.rollback()
            # This create_all will create all tables that have been found in the toolwrapper
            # if not SQLManager.instance().d_database_config['db_connection'] == 'postgresql':
            # TODO: this function is not creating the triggers after the table in postgresql so I switched it off
            IODbPut.create_triggers()
            SQLManager.instance().create_all()
            wrapper_entry.is_content_respected()
        except NoResultFound as e:
            session.rollback()
            raise WopMarsException("Error while parsing the configuration file. The database has not been setUp Correctly.",
                                   str(e))

    def read(self, s_definition_file):
        """
        Reads the file given and insert the rules of the workflow in the database.

        The definition file is supposed to be properly formed. The validation of the content of the definition is done
        during the instanciation of the tools.

        :param: s_definition_file: String containing the path to the definition file.
        :type s_definition_file: str
        :raise: WopmarsException: The content is not validated
        """
        self.load_definition_file(s_definition_file)

        session = SQLManager.instance().get_session()

        # The dict_workflow_definition is assumed to be well formed
        try:
            # The same execution entry for the whole workflow-related database entries.
            execution = Execution(started_at=datetime.datetime.fromtimestamp(time.time()))
            # get the types database entries that should have been created previously
            input_entry = session.query(Type).filter(Type.name == "input").one()
            output_entry = session.query(Type).filter(Type.name == "output").one()
            set_wrapper = set()
            # Encounter a rule block
            for rule in self.__dict_workflow_definition:
                str_wrapper_name = None
                # the name of the rule is extracted after the "rule" keyword. There shouldn't be a ":" but it costs nothing.
                str_rule_name = rule.split()[-1].strip(":")
                Logger.instance().debug("Encounter rule " + str_rule_name + ": \n" +
                                        str(DictUtils.pretty_repr(self.__dict_workflow_definition[rule])))
                # The dict of "input"s, "output"s and "params" is re-initialized for each wrapper
                dict_dict_dict_elm = dict(dict_input={"file": {}, "table": {}},
                                          dict_params={},
                                          dict_output={"file": {}, "table": {}})
                for key_second_step in self.__dict_workflow_definition[rule]:
                    # key_second_step is supposed to be "tool", "input", "output" or "params"
                    if type(self.__dict_workflow_definition[rule][key_second_step]) == dict:
                        # if it is a dict, then inputs, outputs or params are coming
                        for key_third_step in self.__dict_workflow_definition[rule][key_second_step]:
                            # todo tabling modification of the indentation levels + appearance of tables in file
                            if key_second_step == "params":
                                key = key_third_step
                                value = self.__dict_workflow_definition[rule][key_second_step][key_third_step]
                                obj_created = Option(name=key,
                                                     value=value)
                                dict_dict_dict_elm["dict_params"][key] = obj_created
                            else:
                                for key_fourth_step in self.__dict_workflow_definition[rule][key_second_step][key_third_step]:
                                    obj_created = None
                                    if key_third_step == "file":
                                        key = key_fourth_step
                                        str_path_to_file = os.path.join(OptionManager.instance()["--directory"],
                                                                        self.__dict_workflow_definition[rule][
                                                                            key_second_step][
                                                                            key_third_step][
                                                                            key])
                                        obj_created = IOFilePut(name=key,
                                                                path=os.path.abspath(str_path_to_file))

                                    elif key_third_step == "table":
                                        key = key_fourth_step
                                        modelname = self.__dict_workflow_definition[rule][
                                            key_second_step][
                                            key_third_step][
                                            key]
                                        obj_created = IODbPut(model=modelname, tablename=key)

                                        dict_dict_dict_elm["dict_" + key_second_step][
                                            key_third_step][
                                            key] = self.__dict_workflow_definition[rule][key_second_step][key_third_step][key]
                                    # all elements of the current rule block are stored in there
                                    # key_second_step is input or output here
                                    dict_dict_dict_elm["dict_" + key_second_step][key_third_step][key] = obj_created
                                    Logger.instance().debug("Object " + key_second_step + " " + key_third_step + ": " +
                                                            key + " created.")
                    else:
                        # if the step is not a dict, then it is supposed to be the "tool" line
                        str_wrapper_name = self.__dict_workflow_definition[rule][key_second_step]
                # At this point, "dict_dict_dict_elm" is like this:
                # {
                #     'dict_params': {
                #         'option1': Option('option1', 'valueofoption1')
                #     },
                #     'dict_input': {
                #         'file' : {
                #             'input1': IOFilePut('input1', 'path/to/input1')
                #         }
                #         'table': {
                #             'table1': IODbPut('table1', 'package.of.table1')
                #         }
                #     },
                # }

                # Instantiate the refered class and add it to the set of objects
                wrapper_entry = self.create_toolwrapper_entry(str_rule_name, str_wrapper_name, dict_dict_dict_elm, input_entry, output_entry)
                # Associating a toolwrapper to an execution
                wrapper_entry.execution = execution
                set_wrapper.add(wrapper_entry)
                Logger.instance().debug("Object toolwrapper: " + str_wrapper_name + " created.")
                # commit/rollback trick to clean the session - SQLAchemy bug suspected
                session.commit()
                session.rollback()
                # todo set_table_properties outside the rules loop to take into account all the tables at once
                # (error if one tool has a foreign key refering to a table that is not in its I/O put
            IODbPut.set_tables_properties(IODbPut.get_execution_tables())
            session.commit()
            session.rollback()
            # This create_all will create all tables that have been found in the toolwrapper
            # if not SQLManager.instance().d_database_config['db_connection'] == 'postgresql':
            # TODO: this function is not creating the triggers after the table in postgresql so I switched it off
            IODbPut.create_triggers()
            SQLManager.instance().create_all()
            session.add_all(set_wrapper)
            # save all operations done so far.
            session.commit()
            for tw in set_wrapper:
                tw.is_content_respected()

        except NoResultFound as e:
            session.rollback()
            raise WopMarsException("Error while parsing the configuration file. The database has not been setUp Correctly.",
                                   str(e))

    def create_toolwrapper_entry(self, str_rule_name, str_wrapper_name, dict_dict_dict_elm, input_entry, output_entry):
        """
        Actual creating of the Toolwrapper object.

        The toolwrapper object is an entry of the table rule in the resulting database.

        If the scoped_session has current modification, they probably will be commited during this method:
        tables are created and this can only be done with clean session.

        :param str_rule_name: Contains the name of the rule in which the toolwrapper will be used.
        :type str_rule_name: str
        :param str_wrapper_name: Contains the name of the toolwrapper. It will be used for importing the correct module and then for creating the class
        :type str_wrapper_name: str
        :param dict_dict_dict_elm: "input"s "output"s and "params" and will be used to make relations between options / input / output and the toolwrapper.
        :type dict_dict_dict_elm: dict(dict(dict()))
        :param input_entry: input entry
        :type input_entry: :class:`wopmars.framework.bdd.tables.Type.Type`
        :param output_entry: output entry
        :type output_entry: :class:`wopmars.framework.bdd.tables.Type.Type`

        :return: TooLWrapper instance
        """
        session = SQLManager.instance().get_session()
        # Importing the module in the mod variable
        try:
            mod = importlib.import_module(str_wrapper_name)
            # Building the class object
            toolwrapper_class = eval("mod." + str_wrapper_name.split('.')[-1])
        except AttributeError:
            raise WopMarsException("Error while parsing the configuration file: \n\t",
                                   "The class " + str_wrapper_name + " doesn't exist.")
        except ImportError as IE:
            if str_wrapper_name in str(IE):
                raise WopMarsException("Error while parsing the configuration file:",
                                       str_wrapper_name + " module is not in the pythonpath. ")
            else:
                raise WopMarsException("Error while parsing the configuration file:",
                                       str_wrapper_name + " module contains an ImportError: " + str(IE))
        # Initialize the instance of ToolWrapper
        toolwrapper_wrapper = toolwrapper_class(rule_name=str_rule_name)

        # associating ToolWrapper instances with their files / tables
        for elm in dict_dict_dict_elm["dict_input"]:
            if elm == "file":
                for input_f in dict_dict_dict_elm["dict_input"][elm]:
                    # set the type of IOFilePut object
                    iofileput_entry = dict_dict_dict_elm["dict_input"][elm][input_f]
                    iofileput_entry.type = input_entry
                    try:
                        # associating file and toolwrapper
                        toolwrapper_wrapper.files.append(iofileput_entry)
                    except ObjectDeletedError as e:
                        raise WopMarsException("Error in the toolwrapper class declaration. Please, notice the developer",
                                               "The error is probably caused by the lack of the 'polymorphic_identity' attribute"
                                               " in the toolwrapper. Error message: \n" + str(e))
            elif elm == "table":
                for input_t in dict_dict_dict_elm["dict_input"][elm]:
                    # input_t is the name of the table (not the model)
                    # this is a preventing commit because next statement will create a new table and the session has to
                    # be clean. I think it is a bug in SQLAlchemy which not allows queries then insert statements in
                    # the same session
                    session.commit()
                    iodbput_entry = dict_dict_dict_elm["dict_input"][elm][input_t]
                    # the user-side tables are created during the reading of the definition file
                    # table_entry = IODbPut(name=dict_dict_dict_elm["dict_input"][elm][input_t], tablename=input_t)
                    # insert in the database the date of last modification of a developper-side table
                    modification_table_entry, created = session.get_or_create(ModificationTable,
                                                                              defaults={
                                                                                  "date": datetime.datetime.fromtimestamp(
                                                                                      time.time())},
                                                                              table_name=input_t)
                    iodbput_entry.modification = modification_table_entry
                    iodbput_entry.type = input_entry
                    try:
                        toolwrapper_wrapper.tables.append(iodbput_entry)
                    except ObjectDeletedError as e:
                        raise WopMarsException("Error in the toolwrapper class declaration. Please, notice the developer",
                                               "The error is probably caused by the lack of the 'polymorphic_identity' attribute"
                                               " in the toolwrapper. Error message: \n" + str(e))

        for elm in dict_dict_dict_elm["dict_output"]:
            if elm == "file":
                for output_f in dict_dict_dict_elm["dict_output"][elm]:
                    iofileput_entry = dict_dict_dict_elm["dict_output"][elm][output_f]
                    iofileput_entry.type = output_entry
                    try:
                        toolwrapper_wrapper.files.append(iofileput_entry)
                    except ObjectDeletedError as e:
                        raise WopMarsException("Error in the toolwrapper class declaration. Please, notice the developer",
                                               "The error is probably caused by the lack of the 'polymorphic_identity' attribute"
                                               " in the toolwrapper. Error message: \n" + str(e))
            elif elm == "table":
                for output_t in dict_dict_dict_elm["dict_output"][elm]:
                    # output_t is the table name (not the model)
                    session.commit()
                    iodbput_entry = dict_dict_dict_elm["dict_output"][elm][output_t]
                    modification_table_entry, created = session.get_or_create(ModificationTable,
                                                                              defaults={
                                                                                  "date": datetime.datetime.fromtimestamp(
                                                                                      time.time())},
                                                                              table_name=output_t)
                    iodbput_entry.modification = modification_table_entry
                    iodbput_entry.type = output_entry
                    try:
                        toolwrapper_wrapper.tables.append(iodbput_entry)
                    except ObjectDeletedError as e:
                        raise WopMarsException(
                            "Error in the toolwrapper class declaration. Please, notice the developer",
                            "The error is probably caused by the lack of the 'polymorphic_identity' attribute"
                            " in the toolwrapper. Error message: \n" + str(
                                e))

        for opt in dict_dict_dict_elm["dict_params"]:
            # associating option and toolwrapper
            toolwrapper_wrapper.options.append(dict_dict_dict_elm["dict_params"][opt])

        # toolwrapper_wrapper.is_content_respected()
        return toolwrapper_wrapper



"""
Example of module documentation which can be
multiple-lined
"""
from sqlalchemy import Column, Integer, String

from wopmars.framework.database.Base import Base


class FooBaseH(Base):
    """
    Documentation for the class
    """

    __tablename__ = "FooBaseH"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(255))
    state = Column(String)

    __mapper_args__ = {
        'polymorphic_on': state,
        'polymorphic_identity': "1"
    }



import unittest
from unittest import TestCase
from wopmars.framework.database.tables.Option import Option
from wopmars.utils.exceptions.WopMarsException import WopMarsException


class TestOption(TestCase):
    def setUp(self):
        self.__option_string1 = Option(name="option1", value="value1")
        self.__option_string2 = Option(name="option1", value="value1")
        self.__option_string3 = Option(name="option1", value="value2")

        self.__option_int1 = Option(name="option3", value="1")
        self.__option_int2 = Option(name="option3", value="1")
        self.__option_int3 = Option(name="option3", value="2")

        self.__option_float1 = Option(name="option5", value="1.1")
        self.__option_float2 = Option(name="option5", value="1.1")
        self.__option_float3 = Option(name="option5", value="2.2")

        self.__option_bool1 = Option(name="option7", value=True)
        self.__option_bool2 = Option(name="option7", value=True)
        self.__option_bool3 = Option(name="option7", value=False)

    def test_correspond(self):
        try:
            self.__option_string1.correspond("str")
            self.__option_string1.correspond("bool")
        except:
            raise AssertionError("Souldn't raise an exception.")
        self.assertRaises(WopMarsException, self.__option_string1.correspond, "int")
        self.assertRaises(WopMarsException, self.__option_string1.correspond, "float")

        try:
            self.__option_int1.correspond("str")
            self.__option_int1.correspond("int")
            self.__option_int1.correspond("float")
            self.__option_int1.correspond("bool")
        except:
            raise AssertionError("Shouldn't raise an exception.")

        try:
            self.__option_float1.correspond("str")
            self.__option_float1.correspond("float")
            self.__option_float1.correspond("bool")
        except:
            raise AssertionError("Shouldn't raise an exception.")
        self.assertRaises(WopMarsException, self.__option_float1.correspond, "int")

        try:
            self.__option_bool1.correspond("str")
            self.__option_bool1.correspond("int")
            self.__option_bool1.correspond("float")
            self.__option_bool1.correspond("bool")
        except:
            raise AssertionError("Shouldn't raise an exception.")

    def test_eq(self):
        self.assertEqual(self.__option_string1, self.__option_string2)
        self.assertNotEqual(self.__option_string1, self.__option_string3)
        self.assertNotEqual(self.__option_string1, self.__option_int3)

        self.assertEqual(self.__option_int1, self.__option_int2)
        self.assertNotEqual(self.__option_int1, self.__option_int3)
        self.assertNotEqual(self.__option_int1, self.__option_string3)

        self.assertEqual(self.__option_float1, self.__option_float2)
        self.assertNotEqual(self.__option_float1, self.__option_float3)
        self.assertNotEqual(self.__option_float1, self.__option_string3)

        self.assertEqual(self.__option_bool1, self.__option_bool2)
        self.assertNotEqual(self.__option_bool1, self.__option_bool3)
        self.assertNotEqual(self.__option_bool1, self.__option_string3)

if __name__ == "__main__":
    unittest.main()


"""
Module containing the FooWrapper1 class
"""
from wopmars.framework.database.tables.ToolWrapper import ToolWrapper


class Failure(ToolWrapper):
    """
    This class has been done for example/testing purpose.
    Modifications may lead to failure in tests.
    """    
    pass